import pytest

from tensora import Tensor, tensor_method
from tensora.compile import BroadcastTargetIndexError
from tensora.desugar import DiagonalAccessError, NoKernelFoundError


def test_tensor_method():
    A = Tensor.from_aos([(1, 0), (0, 1), (1, 2)], [2.0, -2.0, 4.0], dimensions=(2, 3), format="ds")

    x = Tensor.from_aos([(0,), (1,), (2,)], [3.0, 2.5, 2.0], dimensions=(3,), format="d")

    expected = Tensor.from_aos([(0,), (1,)], [-5.0, 14.0], dimensions=(2,), format="d")

    function = tensor_method("y(i) = A(i,j) * x(j)", {"y": "d", "A": "ds", "x": "d"})

    actual = function(A=A, x=x)

    assert actual == expected


def test_broadcast_target_index_error():
    with pytest.raises(BroadcastTargetIndexError):
        tensor_method("A(i,j) = a(i)", {})


def test_diagonal_error():
    with pytest.raises(DiagonalAccessError):
        tensor_method("a(i) = A(i,i)", {"a": "d", "A": "dd"})


def test_no_solution():
    with pytest.raises(NoKernelFoundError):
        tensor_method("A(i,j) = B(i,j) + C(j,i)", {"A": "ds", "B": "ds", "C": "ds"})



from tensora import Tensor
from tensora import evaluate_taco as evaluate


def test_csr_matrix_vector_product():
    A = Tensor.from_aos([(1, 0), (0, 1), (1, 2)], [2.0, -2.0, 4.0], dimensions=(2, 3), format="ds")

    x = Tensor.from_aos([(0,), (1,), (2,)], [3.0, 2.5, 2.0], dimensions=(3,), format="d")

    expected = Tensor.from_aos([(0,), (1,)], [-5.0, 14.0], dimensions=(2,), format="d")

    actual = evaluate("y(i) = A(i,j) * x(j)", "d", A=A, x=x)

    assert actual == expected


def test_csc_matrix_vector_product():
    A = Tensor.from_aos(
        [(1, 0), (0, 1), (1, 2)], [2.0, -2.0, 4.0], dimensions=(2, 3), format="d1s0"
    )

    x = Tensor.from_aos([(0,), (1,), (2,)], [3.0, 2.5, 2.0], dimensions=(3,), format="d")

    expected = Tensor.from_aos([(0,), (1,)], [-5.0, 14.0], dimensions=(2,), format="d")

    actual = evaluate("y(i) = A(i,j) * x(j)", "d", A=A, x=x)

    assert actual == expected


def test_csr_matrix_plus_csr_matrix():
    A = Tensor.from_aos([(1, 0), (0, 1), (1, 2)], [2.0, -2.0, 4.0], dimensions=(2, 3), format="ds")

    B = Tensor.from_aos([(1, 1), (1, 2), (0, 2)], [-3.0, 4.0, 3.5], dimensions=(2, 3), format="ds")

    expected = Tensor.from_aos(
        [(1, 0), (0, 1), (1, 2), (1, 1), (0, 2)],
        [2.0, -2.0, 8.0, -3.0, 3.5],
        dimensions=(2, 3),
        format="ds",
    )

    actual = evaluate("C(i,j) = A(i,j) + B(i,j)", "ds", A=A, B=B)

    assert actual == expected



from ._cffi_ownership import (
    allocate_taco_structure,
    taco_structure_to_cffi,
    take_ownership_of_arrays,
    take_ownership_of_tensor,
    take_ownership_of_tensor_members,
    tensor_cdefs,
)
from ._initialize_llvm import target
from ._porcelain import evaluate, evaluate_cffi, evaluate_taco, evaluate_tensora, tensor_method
from ._tensor_method import (
    BackendCompiler,
    BroadcastTargetIndexError,
    TensorMethod,
    UnsupportedBackendCompilerError,
)



__all__ = ["parse_format", "parse_named_format"]

from parsita import ParseError, ParserContext, lit, reg, rep
from parsita.util import constant
from returns import result

from ._exceptions import InvalidModeOrderingError
from ._format import Format, Mode


def make_format_with_orderings(dims):
    modes = []
    orderings = []
    for mode, ordering in dims:
        modes.append(mode)
        orderings.append(ordering)
    return Format(tuple(modes), tuple(orderings))


class FormatParsers(ParserContext):
    integer = reg(r"[0-9]+") > int
    dense = lit("d") > constant(Mode.dense)
    compressed = lit("s") > constant(Mode.compressed)
    mode = dense | compressed

    format_without_orderings = rep(mode) > (
        lambda modes: Format(tuple(modes), tuple(range(len(modes))))
    )
    format_with_orderings = rep(mode & integer) > make_format_with_orderings

    format = format_without_orderings | format_with_orderings

    variable = reg(r"[a-zA-Z_][a-zA-Z0-9_]*")
    named_format = variable << ":" & format > tuple


def parse_format(string: str, /) -> result.Result[Format, ParseError | InvalidModeOrderingError]:
    try:
        return FormatParsers.format.parse(string)
    except InvalidModeOrderingError as e:
        return result.Failure(e)


def parse_named_format(
    string: str, /
) -> result.Result[tuple[str, Format], ParseError | InvalidModeOrderingError]:
    try:
        return FormatParsers.named_format.parse(string)
    except InvalidModeOrderingError as e:
        return result.Failure(e)



__all__ = ["Expression", "Literal", "Integer", "Float", "Tensor", "Add", "Multiply"]

from dataclasses import dataclass

from ...format import Mode


class Expression:
    __slots__ = ()


class Literal(Expression):
    __slots__ = ()


@dataclass(frozen=True, slots=True)
class Integer(Literal):
    value: int


@dataclass(frozen=True, slots=True)
class Float(Literal):
    value: float


@dataclass(frozen=True, slots=True)
class Tensor(Expression):
    id: str
    name: str
    indexes: tuple[str, ...]
    modes: tuple[Mode, ...]

    @property
    def order(self):
        return len(self.indexes)


@dataclass(frozen=True, slots=True)
class Add(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class Multiply(Expression):
    left: Expression
    right: Expression



__all__ = ["compile_module"]

import llvmlite.binding as llvm

from ..codegen import ir_to_llvm
from ..ir.ast import Module
from ._initialize_llvm import target


def compile_module(module: Module) -> llvm.ExecutionEngine:
    llvm_ir = ir_to_llvm(module)

    # Compile the module
    llvm_module = llvm.parse_assembly(str(llvm_ir))
    llvm_module.verify()

    # Create target machine
    # We have to recreate this for every module because create_mcjit_compiler
    # takes ownership of target_machine and frees it when the engine goes out of scope
    target_machine = target.create_target_machine()

    # Create execution engine
    backing_mod = llvm.parse_assembly("")
    engine = llvm.create_mcjit_compiler(backing_mod, target_machine)

    # Add the module to the engine and make sure it is ready for execution
    engine.add_module(llvm_module)
    engine.finalize_object()
    engine.run_static_constructors()

    return engine



from __future__ import annotations

__all__ = ["SourceBuilder"]

from abc import abstractmethod
from contextlib import contextmanager
from typing import Mapping

from .._stable_set import StableSet
from .ast import (
    Block,
    Branch,
    Declaration,
    Expression,
    FunctionDefinition,
    Loop,
    Statement,
    Variable,
)
from .types import Type


class Builder:
    def __init__(self):
        self.lines: list[Statement] = []

    @abstractmethod
    def finalize(self):
        raise NotImplementedError()


class BlockBuilder(Builder):
    def __init__(self, comment: str | None = None):
        super().__init__()
        self._comment = comment

    def finalize(self):
        return Block(self.lines, self._comment)


class BranchBuilder(Builder):
    def __init__(self, condition: Expression):
        super().__init__()
        self._condition = condition

    def finalize(self):
        return Branch(self._condition, Block(self.lines), Block([]))


class LoopBuilder(Builder):
    def __init__(self, condition: Expression):
        super().__init__()
        self._condition = condition

    def finalize(self):
        return Loop(self._condition, Block(self.lines))


class FunctionDefinitionBuilder(Builder):
    def __init__(self, name: str, parameters: Mapping[str, Type], return_type: Type):
        super().__init__()
        self._name = name
        self._parameters = parameters
        self._return_type = return_type

    def finalize(self):
        return FunctionDefinition(
            Variable(self._name),
            [Declaration(Variable(name), type) for name, type in self._parameters.items()],
            self._return_type,
            Block(self.lines),
        )


class SourceBuilder:
    def __init__(self, comment: str | None = None):
        self._dependencies: StableSet[str] = StableSet()
        self._stack: list[Builder] = [BlockBuilder(comment)]

    def append(self, statement: Statement | SourceBuilder):
        match statement:
            case Statement():
                self._stack[-1].lines.append(statement)
            case SourceBuilder():
                for dependency in statement._dependencies:
                    self.add_dependency(dependency)
                statement = statement.finalize()
                if statement.comment is not None:
                    self._stack[-1].lines.append(statement)
                else:
                    # This is not simplified by peephole, which only simplifies empty blocks;
                    # it does not inline not blocks with no comments. This is because blocks
                    # with no comments still get newlines between them. This means that
                    # appending a SourceBuilder appends the lines, but appending a block
                    # appends the block itself.
                    self._stack[-1].lines.extend(statement.statements)

    def add_dependency(self, name: str):
        self._dependencies.add(name)

    @contextmanager
    def block(self, comment: str | None = None):
        self._stack.append(BlockBuilder(comment))
        yield None
        self.append(self._stack.pop().finalize())

    @contextmanager
    def branch(self, condition: Expression):
        self._stack.append(BranchBuilder(condition))
        yield None
        self.append(self._stack.pop().finalize())

    @contextmanager
    def loop(self, condition: Expression):
        self._stack.append(LoopBuilder(condition))
        yield None
        self.append(self._stack.pop().finalize())

    @contextmanager
    def function_definition(self, name: str, parameters: Mapping[str, Type], return_type: Type):
        self._stack.append(FunctionDefinitionBuilder(name, parameters, return_type))
        yield None
        self.append(self._stack.pop().finalize())

    def finalize(self) -> Block:
        assert len(self._stack) == 1
        return self._stack[0].finalize()



from __future__ import annotations

__all__ = ["Output"]

from abc import abstractmethod

from ...ir import SourceBuilder
from ...ir.ast import Expression
from ...kernel_type import KernelType
from ..identifiable_expression import TensorLayer


class Output:
    __slots__ = ()

    @abstractmethod
    def write_assignment(
        self, right_hand_side: Expression, kernel_type: KernelType
    ) -> SourceBuilder:
        raise NotImplementedError()

    @abstractmethod
    def next_output(
        self, iteration_output: TensorLayer | None, kernel_type: KernelType
    ) -> tuple[Output, SourceBuilder, SourceBuilder]:
        raise NotImplementedError()



import pytest

from tensora.ir import peephole, peephole_function_definition, peephole_statement
from tensora.ir.ast import *
from tensora.ir.types import *

changed = [
    # Add zero
    (Add(IntegerLiteral(0), Variable("x")), Variable("x")),
    (Add(Variable("x"), IntegerLiteral(0)), Variable("x")),
    (Add(FloatLiteral(0.0), Variable("x")), Variable("x")),
    (Add(Variable("x"), FloatLiteral(0.0)), Variable("x")),
    # Minus zero
    (Subtract(Variable("x"), IntegerLiteral(0)), Variable("x")),
    (Subtract(Variable("x"), FloatLiteral(0.0)), Variable("x")),
    # Multiply 0
    (Multiply(IntegerLiteral(0), Variable("x")), IntegerLiteral(0)),
    (Multiply(Variable("x"), IntegerLiteral(0)), IntegerLiteral(0)),
    (Multiply(FloatLiteral(0.0), Variable("x")), FloatLiteral(0.0)),
    (Multiply(Variable("x"), FloatLiteral(0.0)), FloatLiteral(0.0)),
    # Multiply 1
    (Multiply(IntegerLiteral(1), Variable("x")), Variable("x")),
    (Multiply(Variable("x"), IntegerLiteral(1)), Variable("x")),
    (Multiply(FloatLiteral(1.0), Variable("x")), Variable("x")),
    (Multiply(Variable("x"), FloatLiteral(1.0)), Variable("x")),
    # Equals same
    (Equal(Variable("x"), Variable("x")), BooleanLiteral(True)),
    (LessThanOrEqual(Variable("x"), Variable("x")), BooleanLiteral(True)),
    (GreaterThanOrEqual(Variable("x"), Variable("x")), BooleanLiteral(True)),
    # Not equal same
    (NotEqual(Variable("x"), Variable("x")), BooleanLiteral(False)),
    (LessThan(Variable("x"), Variable("x")), BooleanLiteral(False)),
    (GreaterThan(Variable("x"), Variable("x")), BooleanLiteral(False)),
    # And true
    (And(BooleanLiteral(True), Variable("x")), Variable("x")),
    (And(Variable("x"), BooleanLiteral(True)), Variable("x")),
    # And false
    (And(BooleanLiteral(False), Variable("x")), BooleanLiteral(False)),
    (And(Variable("x"), BooleanLiteral(False)), BooleanLiteral(False)),
    # Or true
    (Or(BooleanLiteral(True), Variable("x")), BooleanLiteral(True)),
    (Or(Variable("x"), BooleanLiteral(True)), BooleanLiteral(True)),
    # Or false
    (Or(BooleanLiteral(False), Variable("x")), Variable("x")),
    (Or(Variable("x"), BooleanLiteral(False)), Variable("x")),
    # Boolean cast constant
    (BooleanToInteger(BooleanLiteral(False)), IntegerLiteral(0)),
    (BooleanToInteger(BooleanLiteral(True)), IntegerLiteral(1)),
    # Branch true
    (Branch(BooleanLiteral(True), Variable("x"), Variable("y")), Variable("x")),
    # Branch false
    (Branch(BooleanLiteral(False), Variable("x"), Variable("y")), Variable("y")),
    # Loop false
    (Loop(BooleanLiteral(False), Variable("x")), Block([])),
    # Empty block
    (Block([Block([], "comment"), Return(IntegerLiteral(0))]), Block([Return(IntegerLiteral(0))])),
    (Branch(Variable("t"), Block([], "comment1"), Block([], "comment2")), Block([])),
    (Loop(Variable("t"), Block([], "comment")), Block([])),
    # Redundant assignment
    (Assignment(Variable("x"), Variable("x")), Block([])),
    (
        Assignment(
            ArrayIndex(Variable("x"), IntegerLiteral(0)),
            ArrayIndex(Variable("x"), IntegerLiteral(0)),
        ),
        Block([]),
    ),
    # Pass through
    (
        ArrayIndex(
            ArrayIndex(Variable("x"), Add(IntegerLiteral(0), Variable("i"))), Variable("j")
        ),
        ArrayIndex(ArrayIndex(Variable("x"), Variable("i")), Variable("j")),
    ),
    (
        ArrayIndex(Variable("x"), Add(IntegerLiteral(0), Variable("i"))),
        ArrayIndex(Variable("x"), Variable("i")),
    ),
    (
        AttributeAccess(ArrayIndex(Variable("x"), Add(IntegerLiteral(0), Variable("i"))), "modes"),
        AttributeAccess(ArrayIndex(Variable("x"), Variable("i")), "modes"),
    ),
    (
        BooleanToInteger(And(BooleanLiteral(True), Variable("t"))),
        BooleanToInteger(Variable("t")),
    ),
    (
        ArrayAllocate(float, Add(IntegerLiteral(0), Variable("x"))),
        ArrayAllocate(float, Variable("x")),
    ),
    (
        ArrayReallocate(Variable("x"), float, Add(IntegerLiteral(0), Variable("y"))),
        ArrayReallocate(Variable("x"), float, Variable("y")),
    ),
    (
        Assignment(Variable("x"), Add(IntegerLiteral(0), Variable("y"))),
        Assignment(Variable("x"), Variable("y")),
    ),
    (
        Assignment(
            ArrayIndex(Variable("x"), Add(IntegerLiteral(0), Variable("i"))), Variable("y")
        ),
        Assignment(ArrayIndex(Variable("x"), Variable("i")), Variable("y")),
    ),
    (
        DeclarationAssignment(
            Declaration(Variable("x"), float), Add(FloatLiteral(0.0), Variable("y"))
        ),
        DeclarationAssignment(Declaration(Variable("x"), float), Variable("y")),
    ),
    (
        Block(
            [
                Assignment(
                    Variable("x"), Add(Add(IntegerLiteral(0), Variable("x")), IntegerLiteral(1))
                )
            ],
            "test block",
        ),
        Block([Assignment(Variable("x"), Add(Variable("x"), IntegerLiteral(1)))], "test block"),
    ),
    (
        Branch(Variable("t"), Variable("x"), Add(IntegerLiteral(0), Variable("y"))),
        Branch(Variable("t"), Variable("x"), Variable("y")),
    ),
    (
        Branch(
            Variable("t"),
            Add(IntegerLiteral(0), Variable("x")),
            Add(IntegerLiteral(0), Variable("y")),
        ),
        Branch(Variable("t"), Variable("x"), Variable("y")),
    ),
    (
        Branch(
            And(BooleanLiteral(True), Variable("t")),
            Variable("x"),
            Add(IntegerLiteral(0), Variable("y")),
        ),
        Branch(Variable("t"), Variable("x"), Variable("y")),
    ),
    (
        Loop(Variable("t"), Add(IntegerLiteral(0), Variable("x"))),
        Loop(Variable("t"), Variable("x")),
    ),
    (
        Loop(And(BooleanLiteral(True), Variable("t")), Variable("x")),
        Loop(Variable("t"), Variable("x")),
    ),
    (
        Return(Add(IntegerLiteral(0), Variable("x"))),
        Return(Variable("x")),
    ),
]


@pytest.mark.parametrize(("before", "after"), changed)
def test_peephole_statement(before: Statement, after: Statement):
    assert peephole_statement(before) == after


left_right_classes = [
    Add,
    Subtract,
    Multiply,
    Equal,
    NotEqual,
    LessThan,
    LessThanOrEqual,
    GreaterThan,
    GreaterThanOrEqual,
    And,
    Or,
    Max,
    Min,
]


@pytest.mark.parametrize("cls", left_right_classes)
def test_pass_through_left_right(cls):
    left = Add(IntegerLiteral(0), Variable("x"))
    right = Add(IntegerLiteral(0), Variable("y"))
    expected = cls(Variable("x"), Variable("y"))
    assert peephole_statement(cls(left, Variable("y"))) == expected
    assert peephole_statement(cls(Variable("x"), right)) == expected


unchanged = [
    Subtract(IntegerLiteral(0), Variable("x")),
    Subtract(FloatLiteral(0.0), Variable("x")),
    Equal(Variable("x"), Variable("y")),
    LessThanOrEqual(Variable("x"), Variable("y")),
    GreaterThanOrEqual(Variable("x"), Variable("y")),
    NotEqual(Variable("x"), Variable("y")),
    LessThan(Variable("x"), Variable("y")),
    GreaterThan(Variable("x"), Variable("y")),
    Loop(BooleanLiteral(True), Variable("x")),
    Assignment(Variable("x"), ArrayIndex(ArrayIndex(Variable("y"), Variable("i")), Variable("j"))),
    Declaration(Variable("x"), float),
]


@pytest.mark.parametrize("input", unchanged)
def test_peephole_statement_noop(input: Statement):
    assert peephole_statement(input) == input


def test_peephole_function_definition():
    function = FunctionDefinition(
        Variable("f"),
        [Declaration(Variable("x"), tensor)],
        integer,
        Return(Multiply(IntegerLiteral(0), IntegerLiteral(1))),
    )
    expected = FunctionDefinition(
        Variable("f"),
        [Declaration(Variable("x"), tensor)],
        integer,
        Return(IntegerLiteral(0)),
    )
    assert peephole_function_definition(function) == expected


def test_peephole_module():
    input_function = FunctionDefinition(
        Variable("f"),
        [Declaration(Variable("x"), tensor)],
        integer,
        Return(Multiply(IntegerLiteral(0), IntegerLiteral(1))),
    )
    expected_function = FunctionDefinition(
        Variable("f"),
        [Declaration(Variable("x"), tensor)],
        integer,
        Return(IntegerLiteral(0)),
    )

    module = Module([input_function, input_function])
    expected = Module([expected_function, expected_function])
    assert peephole(module) == expected



__all__ = ["TensorLayer"]

from dataclasses import dataclass

from ...ir import ast as ir
from .._names import (
    crd_capacity_name,
    crd_name,
    layer_begin_name,
    layer_pointer,
    pos_capacity_name,
    pos_name,
    previous_layer_pointer,
    sparse_end_name,
    vals_capacity_name,
    vals_name,
    value_from_crd,
)
from . import ast as id


@dataclass(frozen=True, slots=True)
class TensorLayer:
    tensor: id.Tensor
    layer: int

    @property
    def mode(self):
        return self.tensor.modes[self.layer]

    def pos_name(self) -> ir.Variable:
        return pos_name(self.tensor.name, self.layer)

    def crd_name(self) -> ir.Variable:
        return crd_name(self.tensor.name, self.layer)

    def vals_name(self) -> ir.Variable:
        return vals_name(self.tensor.name)

    def pos_capacity_name(self) -> ir.Variable:
        return pos_capacity_name(self.tensor.name, self.layer)

    def crd_capacity_name(self) -> ir.Variable:
        return crd_capacity_name(self.tensor.name, self.layer)

    def vals_capacity_name(self) -> ir.Variable:
        return vals_capacity_name(self.tensor.name)

    def layer_pointer(self) -> ir.Variable:
        return layer_pointer(self.tensor.id, self.layer)

    def previous_layer_pointer(self) -> ir.Expression:
        return previous_layer_pointer(self.tensor.id, self.layer)

    def sparse_end_name(self) -> ir.Variable:
        return sparse_end_name(self.tensor.id, self.layer)

    def layer_begin_name(self) -> ir.Variable:
        return layer_begin_name(self.tensor.id, self.layer)

    def value_from_crd(self) -> ir.Variable:
        return value_from_crd(self.tensor.id, self.layer)



import pytest

from tensora.desugar import ast as desugar
from tensora.desugar import desugar_assignment
from tensora.expression import ast as sugar


@pytest.mark.parametrize(
    ("expression", "expected"),
    [
        (
            sugar.Assignment(
                sugar.Tensor("a", ("i",)),
                sugar.Add(sugar.Tensor("b", ("i",)), sugar.Tensor("c", ("i",))),
            ),
            desugar.Assignment(
                desugar.Tensor(0, "a", ("i",)),
                desugar.Add(desugar.Tensor(1, "b", ("i",)), desugar.Tensor(2, "c", ("i",))),
            ),
        ),
        (
            sugar.Assignment(
                sugar.Tensor("A", ("i", "j")),
                sugar.Multiply(sugar.Tensor("B", ("i", "k")), sugar.Tensor("C", ("k", "i"))),
            ),
            desugar.Assignment(
                desugar.Tensor(0, "A", ("i", "j")),
                desugar.Contract(
                    "k",
                    desugar.Multiply(
                        desugar.Tensor(1, "B", ("i", "k")),
                        desugar.Tensor(2, "C", ("k", "i")),
                    ),
                ),
            ),
        ),
        (
            sugar.Assignment(
                sugar.Tensor("A", ("i", "j")),
                sugar.Add(
                    sugar.Tensor("K", ("i", "j")),
                    sugar.Multiply(sugar.Tensor("B", ("i", "k")), sugar.Tensor("C", ("k", "i"))),
                ),
            ),
            desugar.Assignment(
                desugar.Tensor(0, "A", ("i", "j")),
                desugar.Add(
                    desugar.Tensor(1, "K", ("i", "j")),
                    desugar.Contract(
                        "k",
                        desugar.Multiply(
                            desugar.Tensor(2, "B", ("i", "k")),
                            desugar.Tensor(3, "C", ("k", "i")),
                        ),
                    ),
                ),
            ),
        ),
        (
            sugar.Assignment(sugar.Tensor("a", ("i",)), sugar.Tensor("b", ("i", "j"))),
            desugar.Assignment(
                desugar.Tensor(0, "a", ("i",)),
                desugar.Contract("j", desugar.Tensor(1, "b", ("i", "j"))),
            ),
        ),
    ],
)
def test_desugar(expression, expected):
    actual = desugar_assignment(expression)
    assert actual == expected



import hypothesis.strategies as st
from hypothesis import given
from parsita import ParseError
from returns.result import Failure, Success

from tensora.expression import (
    InconsistentDimensionsError,
    MutatingAssignmentError,
    NameConflictError,
    ast,
    parse_assignment,
)
from tensora.format import Format, InvalidModeOrderingError, parse_format

from .strategies import assignments, formats


@given(st.text())
def test_format_parsing_cannot_crash(string):
    match parse_format(string):
        case Success(Format()):
            pass
        case Failure(ParseError() | InvalidModeOrderingError()):
            pass
        case _:
            raise RuntimeError("Unexpected result")


@given(formats())
def test_format_parsing_round_trips(format):
    text = format.deparse()
    new_format = parse_format(text).unwrap()
    assert format == new_format


@given(st.text())
def test_expression_parsing_cannot_crash(string):
    match parse_assignment(string):
        case Success(ast.Assignment()):
            pass
        case Failure(
            ParseError()
            | MutatingAssignmentError()
            | InconsistentDimensionsError()
            | NameConflictError()
        ):
            pass
        case _:
            raise RuntimeError("Unexpected result")


@given(assignments())
def test_expression_parsing_round_trips(assignment):
    text = assignment.deparse()
    new_assignment = parse_assignment(text).unwrap()
    assert assignment == new_assignment



__all__ = ["to_identifiable"]

from ..format import Format
from ..iteration_graph.identifiable_expression import ast as id
from . import ast as desugar


def to_identifiable(self: desugar.Tensor, formats: dict[str, Format]) -> id.Tensor:
    format = formats[self.name]
    return id.Tensor(
        f"{self.id}_{self.name}",
        self.name,
        tuple(self.indexes[i_index] for i_index in format.ordering),
        format.modes,
    )



import pickle

import pytest

from tensora import Format, Mode, Tensor

# The example tensors from the original taco paper:
# http://tensor-compiler.org/kjolstad-oopsla17-tensor-compiler.pdf
figure_5_taco_data = [
    ("d0d1", [[], []], [6, 0, 9, 8, 0, 0, 0, 0, 5, 0, 0, 7]),
    ("s0d1", [[[0, 2], [0, 2]], []], [6, 0, 9, 8, 5, 0, 0, 7]),
    ("d0s1", [[], [[0, 3, 3, 5], [0, 2, 3, 0, 3]]], [6, 9, 8, 5, 7]),
    ("s0s1", [[[0, 2], [0, 2]], [[0, 3, 5], [0, 2, 3, 0, 3]]], [6, 9, 8, 5, 7]),
    ("d1d0", [[], []], [6, 0, 5, 0, 0, 0, 9, 0, 0, 8, 0, 7]),
    ("s1d0", [[[0, 3], [0, 2, 3]], []], [6, 0, 5, 9, 0, 0, 8, 0, 7]),
    ("d1s0", [[], [[0, 2, 2, 3, 5], [0, 2, 0, 0, 2]]], [6, 5, 9, 8, 7]),
    ("s1s0", [[[0, 3], [0, 2, 3]], [[0, 2, 3, 5], [0, 2, 0, 0, 2]]], [6, 5, 9, 8, 7]),
]


@pytest.mark.parametrize(("format", "indices", "vals"), figure_5_taco_data)
def test_figure_5(format, indices, vals):
    data = [[6, 0, 9, 8], [0, 0, 0, 0], [5, 0, 0, 7]]
    A = Tensor.from_lol(data, dimensions=(3, 4), format=format)

    assert A.taco_indices == indices
    assert A.taco_vals == vals


@pytest.mark.parametrize(("format", "indices", "vals"), figure_5_taco_data)
@pytest.mark.parametrize(
    "permutation",
    [
        [0, 1, 2, 3, 4],
        [4, 3, 2, 1, 0],
        [0, 4, 3, 2, 1],
        [2, 1, 3, 4, 0],
    ],
)
def test_unsorted_coordinates(format, indices, vals, permutation):
    data = [
        ((0, 0), 6),
        ((0, 2), 9),
        ((0, 3), 8),
        ((2, 0), 5),
        ((2, 3), 7),
    ]

    permutated_data = [data[i] for i in permutation]
    coordinates, values = zip(*permutated_data, strict=False)

    A = Tensor.from_aos(coordinates, values, dimensions=(3, 4), format=format)

    assert A.taco_indices == indices
    assert A.taco_vals == vals

    A = Tensor.from_soa(zip(*coordinates, strict=False), values, dimensions=(3, 4), format=format)

    assert A.taco_indices == indices
    assert A.taco_vals == vals

    A = Tensor.from_dok(dict(permutated_data), dimensions=(3, 4), format=format)

    assert A.taco_indices == indices
    assert A.taco_vals == vals


# The example tensors from the follow-up taco paper:
# https://tensor-compiler.org/chou-oopsla18-taco-formats.pdf
@pytest.mark.parametrize(
    ("format", "indices", "vals"),
    [("d", [[]], [5, 1, 0, 0, 2, 0, 8, 0]), ("s", [[[0, 4], [0, 1, 4, 6]]], [5, 1, 2, 8])],
)
def test_figure_2a(format, indices, vals):
    data = [5, 1, 0, 0, 2, 0, 8, 0]
    a = Tensor.from_lol(data, dimensions=(8,), format=format)

    assert a.taco_indices == indices
    assert a.taco_vals == vals


@pytest.mark.parametrize(
    ("format", "indices", "vals"),
    [
        ("ds", [[], [[0, 2, 4, 4, 7], [0, 1, 0, 1, 0, 3, 4]]], [5, 1, 7, 3, 8, 4, 9]),
        (
            "ss",
            [[[0, 3], [0, 1, 3]], [[0, 2, 4, 7], [0, 1, 0, 1, 0, 3, 4]]],
            [5, 1, 7, 3, 8, 4, 9],
        ),
    ],
)
def test_figure_2e(format, indices, vals):
    data = {
        (0, 0): 5,
        (0, 1): 1,
        (1, 0): 7,
        (1, 1): 3,
        (3, 0): 8,
        (3, 3): 4,
        (3, 4): 9,
    }
    a = Tensor.from_dok(data, dimensions=(4, 6), format=format)

    assert a.taco_indices == indices
    assert a.taco_vals == vals


@pytest.mark.parametrize(
    ("format", "indices", "vals"),
    [
        (
            "sss",
            [
                [[0, 2], [0, 2]],
                [[0, 2, 5], [0, 2, 0, 2, 3]],
                [[0, 2, 3, 4, 6, 8], [0, 1, 1, 1, 0, 1, 0, 1]],
            ],
            [1, 7, 5, 2, 4, 8, 3, 9],
        ),
    ],
)
def test_figure_2m(format, indices, vals):
    data = {
        (0, 0, 0): 1,
        (0, 0, 1): 7,
        (0, 2, 1): 5,
        (2, 0, 1): 2,
        (2, 2, 0): 4,
        (2, 2, 1): 8,
        (2, 3, 0): 3,
        (2, 3, 1): 9,
    }
    a = Tensor.from_dok(data, dimensions=(3, 4, 2), format=format)

    assert a.taco_indices == indices
    assert a.taco_vals == vals


def test_from_dok():
    data = {
        (2, 2): 2.0,
        (0, 2): -3.0,
        (1, 0): 2.0,
        (2, 3): 5.0,
    }
    format = Format((Mode.compressed, Mode.compressed), (0, 1))
    x = Tensor.from_dok(data, dimensions=(4, 5), format="ss")

    assert x.order == 2
    assert x.dimensions == (4, 5)
    assert x.modes == (Mode.compressed, Mode.compressed)
    assert x.mode_ordering == (0, 1)
    assert x.format == format
    assert x.to_dok() == data


def test_from_aos():
    format = Format((Mode.dense, Mode.compressed), (1, 0))
    x = Tensor.from_aos(
        [(1, 0), (1, 1), (2, 1), (3, 1)],
        [4.5, 3.2, -3.0, 5.0],
        dimensions=(4, 3),
        format=format,
    )

    assert x.order == 2
    assert x.dimensions == (4, 3)
    assert x.modes == (Mode.dense, Mode.compressed)
    assert x.mode_ordering == (1, 0)
    assert x.format == format
    assert x.to_dok() == {
        (1, 0): 4.5,
        (1, 1): 3.2,
        (2, 1): -3.0,
        (3, 1): 5.0,
    }


def test_from_soa():
    format = Format((Mode.dense, Mode.compressed, Mode.compressed), (0, 1, 2))
    x = Tensor.from_soa(
        ([0, 1, 1, 0], [0, 0, 1, 1], [0, 1, 2, 1]),
        [4.5, 3.2, -3.0, 5.0],
        dimensions=(2, 3, 3),
        format=format,
    )

    assert x.order == 3
    assert x.dimensions == (2, 3, 3)
    assert x.modes == (Mode.dense, Mode.compressed, Mode.compressed)
    assert x.mode_ordering == (0, 1, 2)
    assert x.format == format
    assert x.to_dok() == {
        (0, 0, 0): 4.5,
        (1, 0, 1): 3.2,
        (1, 1, 2): -3.0,
        (0, 1, 1): 5.0,
    }


def test_from_dense_lil():
    format = Format((Mode.dense, Mode.dense), (0, 1))
    x = Tensor.from_lol(
        [[0, -4.0, 4.5], [0, -3.5, 2.5]],
        dimensions=(2, 3),
        format=format,
    )

    assert x.order == 2
    assert x.dimensions == (2, 3)
    assert x.modes == (Mode.dense, Mode.dense)
    assert x.mode_ordering == (0, 1)
    assert x.format == format
    assert x.to_dok() == {
        (0, 1): -4.0,
        (0, 2): 4.5,
        (1, 1): -3.5,
        (1, 2): 2.5,
    }


def test_from_dense_lil_scalar():
    format = Format((), ())
    x = Tensor.from_lol(2.0, dimensions=(), format=format)

    assert x.order == 0
    assert x.dimensions == ()
    assert x.modes == ()
    assert x.mode_ordering == ()
    assert x.format == format
    assert x.to_dok() == {(): 2.0}


def test_to_dok():
    dok = {
        (2, 3): 2.0,
        (0, 1): 0.0,
        (1, 2): -1.0,
        (0, 3): 0.0,
    }
    dok_no_zeros = {key: value for key, value in dok.items() if value != 0}
    x = Tensor.from_dok(dok)

    assert x.to_dok() == dok_no_zeros
    assert x.to_dok(explicit_zeros=True) == dok


def test_to_float():
    x = Tensor.from_lol(2.0)
    assert float(x) == 2.0


def test_nonscalar_to_float():
    x = Tensor.from_lol([1, 2])
    with pytest.raises(ValueError, match="Can only convert Tensor of order 0 to float"):
        _ = float(x)


@pytest.mark.parametrize(
    ("a", "b", "c"),
    [
        ([0, 0, 1], 3, [3, 3, 4]),
        (3, [0, 0, 1], [3, 3, 4]),
        ([1, 2, 3], [0, 0, 1], [1, 2, 4]),
    ],
)
def test_add(a, b, c):
    if isinstance(a, list):
        a = Tensor.from_lol(a)
    if isinstance(b, list):
        b = Tensor.from_lol(b)
    expected = Tensor.from_lol(c)

    actual = a + b
    assert actual == expected


@pytest.mark.parametrize(
    ("a", "b", "c"),
    [
        ([0, 0, 1], 3, [-3, -3, -2]),
        (3, [0, 0, 1], [3, 3, 2]),
        ([1, 2, 3], [0, 0, 1], [1, 2, 2]),
    ],
)
def test_subtract(a, b, c):
    if isinstance(a, list):
        a = Tensor.from_lol(a)
    if isinstance(b, list):
        b = Tensor.from_lol(b)
    expected = Tensor.from_lol(c)

    actual = a - b
    assert actual == expected


@pytest.mark.parametrize(
    ("a", "b", "c"),
    [
        ([0, 0, 1], 3, [0, 0, 3]),
        (3, [0, 0, 2], [0, 0, 6]),
        ([1, 2, 3], [0, 0, 4], [0, 0, 12]),
    ],
)
def test_multiply(a, b, c):
    if isinstance(a, list):
        a = Tensor.from_lol(a)
    if isinstance(b, list):
        b = Tensor.from_lol(b)
    expected = Tensor.from_lol(c)

    actual = a * b
    assert actual == expected


@pytest.mark.parametrize(
    ("a", "b", "c"),
    [
        ([3, 2, 5], [1, 2, 0], 7),
        ([0, 0, 1], [[0, 0], [4, -1], [0, 2]], [0, 2]),
        ([[0, 0, 1], [2, 2, 3]], [0, -1, 2], [2, 4]),
        ([[0, 0, 1], [2, 2, 3]], [[0, 0], [4, -1], [0, 2]], [[0, 2], [8, 4]]),
    ],
)
def test_matrix_multiply(a, b, c):
    a = Tensor.from_lol(a)
    b = Tensor.from_lol(b)
    c = Tensor.from_lol(c)

    actual = a @ b
    assert actual == c


def test_binary_mismatched_dimensions():
    a = Tensor.from_lol([3, 2, 5])
    b = Tensor.from_lol([1, 2, 0, 4])

    with pytest.raises(ValueError, match="Cannot apply operator +"):
        _ = a + b

    with pytest.raises(ValueError, match="Cannot apply operator -"):
        _ = a - b

    with pytest.raises(ValueError, match="Cannot apply operator *"):
        _ = a * b


@pytest.mark.parametrize("left", [[3, 2, 5], [[0, 1, 0], [0, 2, 0]]])
@pytest.mark.parametrize("right", [[4, 1], [[0, 1, 0], [0, 2, 0]]])
def test_matrix_multiply_mismatched_dimensions(left, right):
    a = Tensor.from_lol(left)
    b = Tensor.from_lol(right)

    with pytest.raises(ValueError, match="Cannot apply operator @"):
        _ = a @ b


def test_matrix_multiply_too_many_dimensions():
    a = Tensor.from_lol([3, 2, 5])
    b = Tensor.from_dok(
        {
            (0, 0, 0): 4.5,
            (1, 0, 1): 3.2,
            (1, 1, 2): -3.0,
            (0, 1, 1): 5.0,
        },
        dimensions=(3, 3, 3),
    )

    with pytest.raises(ValueError, match="Matrix multiply is only defined"):
        _ = a @ b


def test_equality():
    assert Tensor.from_lol([0, 2, 0], format="d") == Tensor.from_lol([0, 2, 0], format="s")
    assert Tensor.from_lol([0, 1, 0], format="d") != Tensor.from_lol([0, 2, 0], format="s")
    assert Tensor.from_lol([0, 1, 0], format="d") != 1


@pytest.mark.parametrize(
    "tensor",
    [
        Tensor.from_dok({}, dimensions=()),
        Tensor.from_dok({(2, 3): 2.0, (0, 1): 0.0, (1, 2): -1.0, (0, 3): 0.0}, dimensions=(2, 4)),
        Tensor.from_dok(
            {(0, 0, 0): 4.5, (1, 0, 1): 3.2, (1, 1, 2): -3.0, (0, 1, 1): 5.0}, dimensions=(3, 3, 3)
        ),
    ],
)
def test_pickle(tensor):
    # Ensure that not only are the tensors equal, but that they also have the same format and explicit zeros, neither of
    # which affects equality
    data = pickle.dumps(tensor)
    reconstituted = pickle.loads(data)

    assert tensor == reconstituted
    assert tensor.format == reconstituted.format
    assert tensor.to_dok(explicit_zeros=True) == reconstituted.to_dok(explicit_zeros=True)


def test_str_repr():
    # Just make sure these run
    a = Tensor.from_lol([0, 2, 1, 0])
    str(a)
    repr(a)



"""Peephole optimizations.

Current optimization implemented:
* add_zero: 0 + expr or expr + 0 => expr
* minus_zero: expr - 0 => expr
* multiply_zero: 0 * expr or expr * 0 => 0
* multiply_one: 1 * expr or expr * 1 => expr
* equal_same: a == a or a <= a or a >= a => true
* not_equal_same: a != a or a < a or a > a => false
* and_true: expr & true or true & expr => expr
* and_false: expr & false or false & expr => false
* or_true: expr | true or true | expr => true
* or_false: expr | false or false | expr => expr
* boolean_cast_constant: cast(false) => 0 or cast(true) => 1
* branch_true: if (true) then block1 else block2 => block1
* branch_false: if (false) then block1 else block2 => block2
* loop_false: while(false) block => {}
* empty_block: empty blocks get deleted in blocks, branches, and loops
* redundant_assignment: a = a => {}
"""

__all__ = ["peephole_function_definition", "peephole_statement", "peephole"]

from dataclasses import replace
from functools import singledispatch

from .ast import (
    Add,
    And,
    ArrayAllocate,
    ArrayIndex,
    ArrayReallocate,
    Assignable,
    Assignment,
    AttributeAccess,
    Block,
    BooleanLiteral,
    BooleanToInteger,
    Branch,
    Declaration,
    DeclarationAssignment,
    Equal,
    Expression,
    FloatLiteral,
    FunctionDefinition,
    GreaterThan,
    GreaterThanOrEqual,
    IntegerLiteral,
    LessThan,
    LessThanOrEqual,
    Loop,
    Max,
    Min,
    Module,
    Multiply,
    NotEqual,
    Or,
    Return,
    Statement,
    Subtract,
    Variable,
)


@singledispatch
def peephole_assignable(self: Assignable) -> Assignable:
    raise NotImplementedError(f"peephole_assignable not implemented for {type(self)}: {self}")


@peephole_assignable.register(Variable)
def peephole_variable(self: Variable) -> Variable:
    return self


@peephole_assignable.register(AttributeAccess)
def peephole_attribute_access(self: AttributeAccess) -> Assignable:
    return replace(self, target=peephole_assignable(self.target))


@peephole_assignable.register(ArrayIndex)
def peephole_array_index(self: ArrayIndex) -> Assignable:
    return ArrayIndex(peephole_assignable(self.target), peephole_expression(self.index))


@singledispatch
def peephole_expression(self: Expression) -> Expression:
    raise NotImplementedError(f"peephole_expression not implemented for {type(self)}: {self}")


@peephole_expression.register(Assignable)
def peephole_expression_assignable(self: Assignable) -> Assignable:
    # Assignables are expressions in their own right
    return peephole_assignable(self)


@peephole_expression.register(IntegerLiteral)
@peephole_expression.register(FloatLiteral)
@peephole_expression.register(BooleanLiteral)
def peephole_noop(self: Expression) -> Expression:
    return self


@peephole_expression.register(Add)
def peephole_add(self: Add) -> Expression:
    left = peephole_expression(self.left)
    right = peephole_expression(self.right)

    if left == IntegerLiteral(0) or left == FloatLiteral(0.0):
        return right
    elif right == IntegerLiteral(0) or right == FloatLiteral(0.0):
        return left
    else:
        return Add(left, right)


@peephole_expression.register(Subtract)
def peephole_subtract(self: Subtract) -> Expression:
    left = peephole_expression(self.left)
    right = peephole_expression(self.right)

    if right == IntegerLiteral(0) or right == FloatLiteral(0.0):
        return left
    else:
        return Subtract(left, right)


@peephole_expression.register(Multiply)
def peephole_multiply(self: Multiply) -> Expression:
    left = peephole_expression(self.left)
    right = peephole_expression(self.right)

    if left == IntegerLiteral(0) or right == IntegerLiteral(0):
        return IntegerLiteral(0)
    elif left == FloatLiteral(0.0) or right == FloatLiteral(0.0):
        return FloatLiteral(0.0)
    elif left == IntegerLiteral(1) or left == FloatLiteral(1.0):
        return right
    elif right == IntegerLiteral(1) or right == FloatLiteral(1.0):
        return left
    else:
        return Multiply(left, right)


@peephole_expression.register(Equal)
@peephole_expression.register(GreaterThanOrEqual)
@peephole_expression.register(LessThanOrEqual)
def peephole_equal(self: Equal) -> Expression:
    left = peephole_expression(self.left)
    right = peephole_expression(self.right)

    if left == right:
        return BooleanLiteral(True)
    else:
        # Use replace so the class is retained
        return replace(self, left=left, right=right)


@peephole_expression.register(NotEqual)
@peephole_expression.register(GreaterThan)
@peephole_expression.register(LessThan)
def peephole_not_equal(self: NotEqual) -> Expression:
    left = peephole_expression(self.left)
    right = peephole_expression(self.right)

    if left == right:
        return BooleanLiteral(False)
    else:
        # Use replace so the class is retained
        return replace(self, left=left, right=right)


@peephole_expression.register(And)
def peephole_and(self: And) -> Expression:
    left = peephole_expression(self.left)
    right = peephole_expression(self.right)

    if left == BooleanLiteral(False) or right == BooleanLiteral(False):
        return BooleanLiteral(False)
    elif left == BooleanLiteral(True):
        return right
    elif right == BooleanLiteral(True):
        return left
    else:
        return And(left, right)


@peephole_expression.register(Or)
def peephole_or(self: Or) -> Expression:
    left = peephole_expression(self.left)
    right = peephole_expression(self.right)

    if left == BooleanLiteral(True) or right == BooleanLiteral(True):
        return BooleanLiteral(True)
    elif left == BooleanLiteral(False):
        return right
    elif right == BooleanLiteral(False):
        return left
    else:
        return Or(left, right)


@peephole_expression.register(Max)
@peephole_expression.register(Min)
def peephole_max_min(self: Max) -> Expression:
    left = peephole_expression(self.left)
    right = peephole_expression(self.right)

    # Use replace so the class is retained
    return replace(self, left=left, right=right)


@peephole_expression.register(BooleanToInteger)
def peephole_boolean_to_integer(self: BooleanToInteger) -> Expression:
    expression = peephole_expression(self.expression)

    if expression == BooleanLiteral(False):
        return IntegerLiteral(0)
    elif expression == BooleanLiteral(True):
        return IntegerLiteral(1)
    else:
        return BooleanToInteger(expression)


@peephole_expression.register(ArrayAllocate)
def peephole_array_allocate(self: ArrayAllocate) -> Expression:
    n_elements = peephole_expression(self.n_elements)
    return replace(self, n_elements=n_elements)


@peephole_expression.register(ArrayReallocate)
def peephole_array_reallocate(self: ArrayReallocate) -> Expression:
    old = peephole_assignable(self.old)
    n_elements = peephole_expression(self.n_elements)
    return replace(self, old=old, n_elements=n_elements)


@singledispatch
def peephole_statement(self: Statement) -> Statement:
    raise NotImplementedError(f"peephole not implemented for {type(self)}: {self}")


@peephole_statement.register(Expression)
def peephole_expression_statement(self: Expression) -> Expression:
    # Expressions are statements in their own right
    return peephole_expression(self)


@peephole_statement.register(Declaration)
def peephole_declaration(self: Declaration) -> Declaration:
    return self


@peephole_statement.register(Assignment)
def peephole_assignment(self: Assignment) -> Statement:
    target = peephole_assignable(self.target)
    value = peephole_expression(self.value)

    if target == value:
        return Block([])
    else:
        return Assignment(target, value)


@peephole_statement.register(DeclarationAssignment)
def peephole_declaration_assignment(self: DeclarationAssignment) -> Statement:
    value = peephole_expression(self.value)
    return replace(self, value=value)


@peephole_statement.register(Block)
def peephole_block(self: Block) -> Statement:
    statements = []
    for old_statement in self.statements:
        statement = peephole_statement(old_statement)
        if isinstance(statement, Block) and statement.is_empty():
            pass
        else:
            statements.append(statement)

    return replace(self, statements=statements)


@peephole_statement.register(Branch)
def peephole_branch(self: Branch) -> Statement:
    condition = peephole_expression(self.condition)
    if_true = peephole_statement(self.if_true)
    if_false = peephole_statement(self.if_false)

    if condition == BooleanLiteral(True):
        return if_true
    elif condition == BooleanLiteral(False):
        return if_false
    elif (
        isinstance(if_true, Block)
        and if_true.is_empty()
        and isinstance(if_false, Block)
        and if_false.is_empty()
    ):
        return Block([])
    else:
        return Branch(condition, if_true, if_false)


@peephole_statement.register(Loop)
def peephole_loop(self: Loop) -> Statement:
    condition = peephole_expression(self.condition)
    body = peephole_statement(self.body)

    if condition == BooleanLiteral(False):
        return Block([])
    elif isinstance(self.body, Block) and self.body.is_empty():
        return Block([])
    else:
        return Loop(condition, body)


@peephole_statement.register(Return)
def peephole_return(self: Return) -> Statement:
    value = peephole_expression(self.value)
    return Return(value)


def peephole_function_definition(self: FunctionDefinition) -> FunctionDefinition:
    body = peephole_statement(self.body)
    return replace(self, body=body)


def peephole(self: Module) -> Module:
    functions = [peephole_function_definition(function) for function in self.definitions]
    return Module(functions)



from __future__ import annotations

__all__ = ["generate_ir"]

from functools import singledispatch

from ..format import Mode
from ..ir import SourceBuilder, types
from ..ir.ast import (
    And,
    Block,
    BooleanToInteger,
    Branch,
    Declaration,
    Equal,
    Expression,
    FunctionDefinition,
    GreaterThan,
    IntegerLiteral,
    LessThan,
    Min,
    Return,
    Variable,
)
from ..kernel_type import KernelType
from ._definition import Definition
from ._names import crd_name, dimension_name, pos_name, vals_name
from ._write_sparse_ir import (
    write_crd_assembly,
    write_pos_allocation,
    write_pos_assembly,
    write_sparse_initialization,
)
from .identifiable_expression import to_ir
from .identifiable_expression._tensor_layer import TensorLayer
from .iteration_graph import IterationGraph, IterationNode, SumNode, TerminalNode
from .outputs import AppendOutput, Output


@singledispatch
def to_ir_iteration_graph(
    self: IterationGraph, output: Output, kernel_type: KernelType
) -> SourceBuilder:
    raise NotImplementedError(
        f"iteration_graph_to_c_code not implemented for {type(self)}: {self}"
    )


@to_ir_iteration_graph.register(TerminalNode)
def to_ir_terminal_expression(self: TerminalNode, output: Output, kernel_type: KernelType):
    source = SourceBuilder("*** Computation of expression ***")

    if kernel_type.is_compute():
        source.append(output.write_assignment(to_ir(self.expression), kernel_type))

    return source


def generate_subgraphs(graph: IterationNode) -> list[IterationNode]:
    # The 0th element is just the full graph
    # Each element is derived from a previous element by zeroing a tensor
    # Zeroing a tensor always results in a strictly simpler graph
    # Zeroing a tensor will cause tensors multiplied by it to be zeroed as well
    # This means that zeroing two different tensors can result in the same graph
    # Subgraphs can be keyed by their set of remaining sparse tensors to eliminate duplicates
    all_subgraphs = {graph.compressed_dimensions(): graph}
    old_subgraphs = all_subgraphs.copy()

    while True:
        new_graphs = {}

        for sparse_layers, old_graph in old_subgraphs.items():
            # Reverse sparse_dimensions so that last values dropped first
            for sparse_layer in reversed(sparse_layers):
                new_graph = old_graph.exhaust_tensor(sparse_layer)
                new_graphs[new_graph.compressed_dimensions()] = new_graph

        if len(new_graphs) == 0:
            break
        else:
            all_subgraphs.update(new_graphs)
            old_subgraphs = new_graphs

    return list(all_subgraphs.values())


@to_ir_iteration_graph.register(IterationNode)
def to_ir_iteration_variable(self: IterationNode, output: Output, kernel_type: KernelType):
    source = SourceBuilder(f"*** Iteration over {self.index_variable} ***")

    if not kernel_type.is_compute() and not self.has_assemble():
        # If not doing compute and no assembly is left, return early. This optimization avoids
        # generating unecessary looping that would make the code look bad (because the peephole
        # optimizer won't remove it) but wouldn't affect performance (because the backend compiler
        # would remove it).
        return source

    loop_variable = Variable(self.index_variable)

    # If this node is sparse, then we can use efficient sparse iteration. Sparse nodes produce
    # sparse subnodes, so sparsity is a constant for this iteration. A node is sparse if its input
    # is sparse and either its output is sparse or it is a contraction.
    is_sparse = self.is_sparse_input() and (self.output is None or self.is_sparse_output())

    # Compute the next output
    next_output, next_output_declarations, next_output_cleanup = output.next_output(
        self.output, kernel_type
    )
    source.append(next_output_declarations)

    ##################
    # Initialization #
    ##################
    if not is_sparse:
        # If this node is dense, we need to initialize the index variable, which will be advanced
        # by each subnode loop. This could be done even if the node were sparse, but it would be
        # dead code since the sparse index advancement would overwrite it. The sparse index
        # advancement is guaranteed to set the index variable to at least zero, but it also could
        # set it to a larger value.
        source.append(loop_variable.declare(types.integer).assign(0))

    for leaf in self.sparse_leaves():
        source.append(write_sparse_initialization(leaf))

    ############
    # Subnodes #
    ############
    # Whether the node is sparse or dense, we need to handle the fact that there may be sparse
    # layers in this computation. We need to extract the next index from each sparse layer to check
    # if it matches the current index variable. Extracting this index is illegal if the layer is
    # exhausted, so we split the node into subnodes, which is all possible combinations of the
    # sparse leaves zeroed out. Each subnode is emitted one after the other, each advancing the
    # iteration, but only while all its leaves are not exhausted. Once any leaf of a subnode is
    # exhausted, the subnode falls through to the next subnode.

    subnodes = generate_subgraphs(self)
    for subnode in subnodes:
        if is_sparse and len(subnode.compressed_dimensions()) == 0:
            # If there are no sparse leaves, then this is the last subnode and we are in one of
            # these four situations:
            # 1. The input and output are both and dense, so we simply finish the loop.
            # 2. The input is sparse, but the output is dense, so we finish the loop, writing zeros
            #    to the dense layer.
            # 3. The input is dense, but the output is sparse, so we finish the loop, writing
            #    explicit zeros to the sparse layer.
            # 4. The input and output are both sparse, so we skip the implicit zero entirely.
            #
            # This last case is handled by this if statement. Because it is guarenteed to the be
            # the last subnode, break or continue would do the same thing.
            continue

        sparse_subnode_leaves = subnode.sparse_leaves()
        dense_subnode_leaves = subnode.dense_leaves()

        ########
        # Loop #
        ########
        if len(sparse_subnode_leaves) > 0:
            # Loop inside this subnode only as long as the all sparse leaves have not been
            # exhausted.
            while_criteria = And.join(
                [
                    LessThan(leaf.layer_pointer(), leaf.sparse_end_name())
                    for leaf in sparse_subnode_leaves
                ]
            )
        else:
            # The last subnode may be dense, so the last loop should finish out the index variable.
            # This condition could be folded into the sparse leaves condition above, but it would
            # be dead code because all the sparse leaves will always be exhausted first. This dead
            # code would not even be optimized out because the backend compiler does not know that
            # all sparse indexes are (or must be) less than the dimension size.
            while_criteria = LessThan(loop_variable, dimension_name(self.index_variable))

        with source.loop(while_criteria):
            ##########################
            # Extract sparse indexes #
            ##########################
            index_variables = []
            for leaf in sparse_subnode_leaves:
                index_variable = leaf.value_from_crd()
                index_variables.append(index_variable)
                source.append(
                    index_variable.declare(types.integer).assign(
                        leaf.crd_name().idx(leaf.layer_pointer())
                    )
                )

            if is_sparse:
                # If the node is sparse, the index variable is the closest value among all the
                # sparse layers.
                source.append(
                    loop_variable.declare(types.integer).assign(Min.join(index_variables))
                )

            #########################
            # Compute dense indexes #
            #########################
            # Earlier layers may not be iterated over yet if they are dense. The position of
            # earlier layers are prerequisites for computing the position of this layer. Defer
            # computing the position of this layer until all earlier layers have been iterated.
            # This means that we must check if later layers have already been iterated over and can
            # now be satisfied. If so, compute the position of those layers now.

            if (
                self.output is not None
                and self.output.mode == Mode.dense
                and isinstance(output, AppendOutput)
            ):
                maybe_dense_output = [self.output]
            else:
                maybe_dense_output = []

            for leaf in maybe_dense_output + dense_subnode_leaves:
                needed_indexes: set[str] = set()
                for i_layer in reversed(range(leaf.layer)):
                    if leaf.tensor.modes[i_layer] == Mode.dense:
                        needed_indexes.add(leaf.tensor.indexes[i_layer])
                    else:
                        break

                layers_to_write = []
                for i_layer in range(leaf.layer, leaf.tensor.order):
                    if leaf.tensor.modes[i_layer] != Mode.dense:
                        # Only consider adjacent dense layers
                        break

                    needed_indexes.add(leaf.tensor.indexes[i_layer])
                    available_before_here = needed_indexes - self.later_indexes()
                    available_after_here = available_before_here | {self.index_variable}

                    if not needed_indexes.issubset(
                        available_before_here
                    ) and needed_indexes.issubset(available_after_here):
                        # Layers whose prerequisites are satisfied with this layer, but not
                        # satisfied without it are written now.
                        layers_to_write.append(TensorLayer(leaf.tensor, i_layer))

                for layer in layers_to_write:
                    pointer = layer.layer_pointer()
                    previous_pointer = layer.previous_layer_pointer()
                    index_variable_i = layer.tensor.indexes[layer.layer]
                    pointer_value = previous_pointer.times(dimension_name(index_variable_i)).plus(
                        index_variable_i
                    )
                    source.append(pointer.declare(types.integer).assign(pointer_value))

            ###############
            # Subsubnodes #
            ###############
            # Before recursing, we need to filter out the sparse leaves with an implicit zero at
            # the current index. We generate all possible combinations of zeroed-out sparse leaves
            # and emit them one after the other in exclusive branches. Each branch is entered on
            # the correct leaves having the current index variable value.

            subsubnodes = generate_subgraphs(subnode)
            subsubnode_leaves: list[tuple[Expression, Block]] = []
            for subsubnode in subsubnodes:
                if is_sparse and len(subsubnode.compressed_dimensions()) == 0:
                    # If there are no sparse leaves, then this is the last subsubnode and we are in
                    # the same four situations as for subnodes above:
                    # 1. The input and output are both and dense, so we write the value.
                    # 2. The input is sparse, but the output is dense, so we write a zero.
                    # 3. The input is dense, but the output is sparse, so we write an explicit
                    #    zero to the sparse layer.
                    # 4. The input and output are both sparse, so we write nothing.
                    #
                    # Like above, this is necessarily the last subsubnode, so break or continue
                    # would do the same thing.
                    continue

                sparse_subsubnode_leaves = subsubnode.sparse_leaves()

                ############################
                # Branch on sparse matches #
                ############################
                condition = And.join(
                    [
                        Equal(leaf.value_from_crd(), loop_variable)
                        for leaf in sparse_subsubnode_leaves
                    ]
                )
                block = SourceBuilder()

                ###################################
                # Allocate space for pos and vals #
                ###################################
                # This is done at this level, rather than up a level, because none of the
                # subsubnode conditions may hit. In that case, no allocation should be done.
                if (
                    kernel_type.is_assemble()
                    and self.is_sparse_output()
                    and isinstance(next_output, AppendOutput)
                ):
                    block.append(write_pos_allocation(self.output))

                ################################
                # Store position of next layer #
                ################################
                # This allows the current sparse layer to remember if the next sparse layer had any
                # nonzeros. This is conceptually part of the next layer, but it is used by this
                # layer's crd assembly, so we put it here. We could constitutively have each layer
                # store a bool indicating whether or not it wrote anything, which woould be
                # optimized out when either layer was not sparse, but the upper layer would still
                # need to probe the lower layer, so not that much would be gained.
                if self.is_sparse_output() and self.next.is_sparse_output():
                    with block.block("Save next layer's position"):
                        next_output_layer: TensorLayer = self.next.output
                        next_pointer = next_output_layer.layer_pointer()
                        next_pointer_begin = next_output_layer.layer_begin_name().declare(
                            types.integer
                        )
                        block.append(next_pointer_begin.assign(next_pointer))

                #####################
                # Invoke next layer #
                #####################
                block.append(to_ir_iteration_graph(subsubnode.next, next_output, kernel_type))

                ########################################################
                # Write sparse output index and advance output pointer #
                ########################################################
                if self.is_sparse_output() and isinstance(next_output, AppendOutput):
                    if self.next.is_sparse_output():
                        # Only advance the index for this sparse layer if the next sparse layer
                        # had any nonzeros.
                        next_output_layer: TensorLayer = self.next.output
                        next_pointer = next_output_layer.layer_pointer()
                        next_pointer_begin = next_output_layer.layer_begin_name()
                        with block.branch(GreaterThan(next_pointer, next_pointer_begin)):
                            if kernel_type.is_assemble():
                                block.append(write_crd_assembly(self.output))
                            block.append(self.output.layer_pointer().increment())
                    else:
                        if kernel_type.is_assemble():
                            block.append(write_crd_assembly(self.output))
                        block.append(self.output.layer_pointer().increment())

                subsubnode_leaves.append((condition, block.finalize()))

            source.append(Branch.join(subsubnode_leaves))

            #######################
            # Increment positions #
            #######################
            # Increment the position of each sparse layer inside the subnode (because that's where
            # the loop is and accessing the position of an exhausted layer is a memory violation)
            # but outside the subsubnode (because all matching layers need to be incremented, even
            # those that were not computed because they were multiplied with a leaf that had a zero
            # at this index)
            for leaf in sparse_subnode_leaves:
                source.append(
                    leaf.layer_pointer().increment(
                        BooleanToInteger(Equal(leaf.value_from_crd(), loop_variable))
                    )
                )

            if not is_sparse:
                # If this node is dense, we need to increment the index variable. This could be
                # done even if the node were sparse, but it would be dead code since the sparse
                # index advancement would overwrite it. The sparse index advancement is guaranteed
                # to advance the index at least one to match this increment, but it also could
                # advance further.
                source.append(loop_variable.increment())

    ################################
    # Write sparse output position #
    ################################
    if (
        kernel_type.is_assemble()
        and self.is_sparse_output()
        and isinstance(next_output, AppendOutput)
    ):
        source.append(write_pos_assembly(self.output))

    source.append(next_output_cleanup)

    return source


@to_ir_iteration_graph.register(SumNode)
def to_ir_sum(self: SumNode, output: Output, kernel_type: KernelType):
    source = SourceBuilder("*** Sum ***")

    if kernel_type.is_compute():
        # No assembly is currently allowed downstream of a Sum node
        next_output, next_output_declarations, next_output_cleanup = output.next_output(
            None, kernel_type
        )
        source.append(next_output_declarations)

        for term in self.terms:
            source.append(to_ir_iteration_graph(term, next_output, kernel_type))

    return source


def generate_ir(
    definition: Definition, graph: IterationGraph, kernel_type: KernelType
) -> FunctionDefinition:
    # Function body
    source = SourceBuilder()

    # Dimensions of all index variables
    with source.block("Extract dimensions"):
        for index_name, tensor_layer in definition.indexes.items():
            declaration = dimension_name(index_name).declare(types.integer)
            value = Variable(tensor_layer.name).attr("dimensions").idx(tensor_layer.dimension)
            source.append(declaration.assign(value))

    # Unpack tensors
    with source.block("Unpack tensors"):
        for tensor_name, format in definition.formats.items():
            for i, mode in enumerate(format.modes):
                match mode:
                    case Mode.dense:
                        pass
                    case Mode.compressed:
                        pos_declaration = pos_name(tensor_name, i).declare(
                            types.Pointer(types.integer)
                        )
                        pos_value = Variable(tensor_name).attr("indices").idx(i).idx(0)
                        source.append(pos_declaration.assign(pos_value))
                        crd_declaration = crd_name(tensor_name, i).declare(
                            types.Pointer(types.integer)
                        )
                        crd_value = Variable(tensor_name).attr("indices").idx(i).idx(1)
                        source.append(crd_declaration.assign(crd_value))

            vals_declaration = vals_name(tensor_name).declare(types.Pointer(types.float))
            vals_value = Variable(tensor_name).attr("vals")
            source.append(vals_declaration.assign(vals_value))

    output = AppendOutput(definition.output_variable, 0)
    source.append(output.write_declarations(kernel_type))

    source.append(to_ir_iteration_graph(graph, output, kernel_type))

    source.append(output.write_cleanup(kernel_type))

    source.append(Return(IntegerLiteral(0)))

    # Function declaration
    parameters = [
        Declaration(Variable(name), types.Pointer(types.tensor))
        for name in definition.formats.keys()
    ]
    function_definition = FunctionDefinition(
        Variable(kernel_type.name), parameters, types.integer, source.finalize()
    )

    return function_definition



__all__ = ["to_iteration_graphs"]

from collections import defaultdict
from dataclasses import replace
from functools import reduce, singledispatch
from itertools import chain, count, permutations, product
from typing import Iterator

from ..format import Format, Mode
from ..iteration_graph import iteration_graph as ig
from ..iteration_graph.identifiable_expression import TensorLayer
from ..iteration_graph.identifiable_expression import ast as id
from . import ast


def legal_iteration_orders(format: Format) -> Iterator[list[int]]:
    """Legal iteration orders for layers of a tensor of a given format.

    Zero is the first layer, not the first index. For example, a ds format has
    one legal iteration order (i.e. [0, 1]) and a d1s0 format has a different
    legal iteration order (i.e. [1, 0])."""
    reorderable_groups = []
    restart = True
    for i, mode in enumerate(format.modes):
        match mode:
            case Mode.dense:
                if restart:
                    reorderable_groups.append([i])
                else:
                    reorderable_groups[-1].append(i)
                restart = False
            case Mode.compressed:
                reorderable_groups.append([i])
                restart = True

    for group_order in product(*(permutations(group) for group in reorderable_groups)):
        yield list(chain(*group_order))


@singledispatch
def to_iteration_graphs_expression(
    self: ast.Expression,
    formats: dict[str, Format],
    counter: Iterator[int],
) -> Iterator[ig.IterationGraph]:
    raise NotImplementedError(
        f"to_iteration_graphs_expression not implemented for {type(self)}: {self}"
    )


@to_iteration_graphs_expression.register(ast.Integer)
def to_iteration_graphs_integer(
    self: ast.Integer,
    formats: dict[str, Format],
    counter: Iterator[int],
) -> Iterator[ig.IterationGraph]:
    yield ig.TerminalNode(id.Integer(self.value))


@to_iteration_graphs_expression.register(ast.Float)
def to_iteration_graphs_float(
    self: ast.Float,
    formats: dict[str, Format],
    counter: Iterator[int],
) -> Iterator[ig.IterationGraph]:
    yield ig.TerminalNode(id.Float(self.value))


@to_iteration_graphs_expression.register(ast.Tensor)
def to_iteration_graphs_tensor(
    self: ast.Tensor,
    formats: dict[str, Format],
    counter: Iterator[int],
) -> Iterator[ig.IterationGraph]:
    from ._exceptions import DiagonalAccessError

    format = formats[self.name]
    index_variables = tuple(self.indexes[i_index] for i_index in format.ordering)
    modes = formats[self.name].modes

    if len(set(index_variables)) != len(index_variables):
        raise DiagonalAccessError(self)

    for index_order in legal_iteration_orders(format):
        graph = ig.TerminalNode(
            id.Tensor(f"{self.id}_{self.name}", self.name, index_variables, modes)
        )
        # Build iteration order bottom up
        for i_index in reversed(index_order):
            index_variable = index_variables[i_index]
            graph = ig.IterationNode(
                index_variable,
                None,
                next=graph,
            )
        yield graph


def simplify_add(graph: ig.SumNode) -> ig.IterationGraph:
    """Simplify an Add by combining terms with the same index variable.

    An Add node can be simplified by combining all its `TerminalExpression`s into a single
    expression and by combinbing all its `IterationVariable`s with the same index variable into a
    single loop.
    """
    # graph.terms are guaranteed to be IterationVariable or TerminalExpression

    terminal_nodes: list[ig.TerminalNode] = []
    iteration_nodes: defaultdict[str, list[ig.IterationNode]] = defaultdict(list)
    for term in graph.terms:
        match term:
            case ig.TerminalNode():
                terminal_nodes.append(term)
            case ig.IterationNode():
                iteration_nodes[term.index_variable].append(term)

    combined_nodes = []

    if len(terminal_nodes) > 0:
        expression = reduce(id.Add, [term.expression for term in terminal_nodes])
        combined_nodes.append(ig.TerminalNode(expression))

    for terms in iteration_nodes.values():
        # Zip the iteration nodes as far as they will go
        head = terms[0]

        next_terms = []
        for term in terms:
            match term.next:
                case ig.SumNode():
                    # Merge sum nodes
                    next_terms.extend(term.next.terms)
                case _:
                    next_terms.append(term.next)

        combined_nodes.append(replace(head, next=simplify_add(ig.SumNode(graph.name, next_terms))))

    match combined_nodes:
        case [single_term]:
            return single_term
        case _:
            return ig.SumNode(graph.name, combined_nodes)


@to_iteration_graphs_expression.register(ast.Add)
def to_iteration_graphs_add(
    self: ast.Add,
    formats: dict[str, Format],
    counter: Iterator[int],
) -> Iterator[ig.IterationGraph]:
    name = f"sum_{next(counter)}"
    for left in to_iteration_graphs_expression(self.left, formats, counter):
        for right in to_iteration_graphs_expression(self.right, formats, counter):
            # Always simplify Add within Add
            match (left, right):
                case (ig.SumNode(), ig.SumNode()):
                    graph = ig.SumNode(name, [*left.terms, *right.terms])
                case (ig.SumNode(), _):
                    graph = ig.SumNode(name, [*left.terms, right])
                case (_, ig.SumNode()):
                    graph = ig.SumNode(name, [left, *right.terms])
                case (_, _):
                    graph = ig.SumNode(name, [left, right])

            yield simplify_add(graph)


def merge_multiply(
    left: ig.IterationGraph, right: ig.IterationGraph
) -> Iterator[ig.IterationGraph]:
    match (left, right):
        case (ig.TerminalNode(), ig.TerminalNode()):
            yield ig.TerminalNode(id.Multiply(left.expression, right.expression))
        case (ig.IterationNode(), ig.TerminalNode()):
            for tail in merge_multiply(left.next, right):
                yield replace(left, next=tail)
        case (ig.TerminalNode(), ig.IterationNode()):
            for tail in merge_multiply(left, right.next):
                yield replace(right, next=tail)
        case (ig.IterationNode(), ig.IterationNode()):
            if left.index_variable == right.index_variable:
                for tail in merge_multiply(left.next, right.next):
                    yield replace(left, next=tail)
            else:
                if left.index_variable not in right.next.later_indexes():
                    for tail in merge_multiply(left.next, right):
                        yield replace(left, next=tail)

                if right.index_variable not in left.next.later_indexes():
                    for tail in merge_multiply(left, right.next):
                        yield replace(right, next=tail)
        case (ig.SumNode(), _):
            for terms in product(*[merge_multiply(term, right) for term in left.terms]):
                yield ig.SumNode(left.name, list(terms))
        case (_, ig.SumNode()):
            for terms in product(*[merge_multiply(left, term) for term in right.terms]):
                yield ig.SumNode(right.name, list(terms))


@to_iteration_graphs_expression.register(ast.Multiply)
def to_iteration_graphs_multiply(
    self: ast.Multiply,
    formats: dict[str, Format],
    counter: Iterator[int],
) -> Iterator[ig.IterationGraph]:
    for left in to_iteration_graphs_expression(self.left, formats, counter):
        for right in to_iteration_graphs_expression(self.right, formats, counter):
            yield from merge_multiply(left, right)


@to_iteration_graphs_expression.register(ast.Contract)
def to_iteration_graphs_contract(
    self: ast.Contract,
    formats: dict[str, Format],
    counter: Iterator[int],
) -> Iterator[ig.IterationGraph]:
    yield from to_iteration_graphs_expression(self.expression, formats, counter)


def merge_assignment(
    target: ig.IterationGraph, expression: ig.IterationGraph, output_layers: dict[str, TensorLayer]
) -> Iterator[ig.IterationGraph]:
    match (target, expression):
        case (ig.TerminalNode(), _):
            yield expression
        case (ig.IterationNode(), ig.TerminalNode()):
            output_leaf = output_layers[target.index_variable]
            for tail in merge_assignment(target.next, expression, output_layers):
                yield replace(target, output=output_leaf, next=tail)
        case (ig.IterationNode(), ig.IterationNode()):
            if target.index_variable == expression.index_variable:
                output_leaf = output_layers[target.index_variable]
                for tail in merge_assignment(target.next, expression.next, output_layers):
                    yield replace(target, output=output_leaf, next=tail)
            else:
                if target.index_variable not in expression.next.later_indexes():
                    output_leaf = output_layers[target.index_variable]
                    for tail in merge_assignment(target.next, expression, output_layers):
                        yield replace(target, output=output_leaf, next=tail)

                if expression.index_variable not in target.next.later_indexes():
                    for tail in merge_assignment(target, expression.next, output_layers):
                        yield replace(expression, next=tail)
        case (ig.IterationNode(), ig.SumNode(name=name, terms=terms)):
            for merged_terms in product(
                *(merge_assignment(target, term, output_layers) for term in terms)
            ):
                yield simplify_add(ig.SumNode(name, list(merged_terms)))


def to_iteration_graphs(
    assignment: ast.Assignment, formats: dict[str, Format]
) -> Iterator[ig.IterationGraph]:
    output_format = formats[assignment.target.name]
    output_layers = {
        assignment.target.indexes[i_dimension]: TensorLayer(
            id.Tensor(
                f"{assignment.target.id}_{assignment.target.name}",
                assignment.target.name,
                tuple(assignment.target.indexes[i] for i in output_format.ordering),
                output_format.modes,
            ),
            i_layer,
        )
        for i_layer, i_dimension in enumerate(output_format.ordering)
    }

    for target_graph in to_iteration_graphs_expression(assignment.target, formats, []):
        for expression_graph in to_iteration_graphs_expression(
            assignment.expression, formats, count(1)
        ):
            for graph in merge_assignment(target_graph, expression_graph, output_layers):
                yield graph



__all__ = ["target"]

import llvmlite.binding as llvm

# Initialize the LLVM
# https://llvmlite.readthedocs.io/en/latest/user-guide/binding/examples.html
llvm.initialize()
llvm.initialize_native_target()
llvm.initialize_native_asmprinter()

# Create the target representing the current host
target = llvm.Target.from_default_triple()



__all__ = ["best_algorithm"]

from returns.result import Failure, Result, Success

from ..format import Format
from ..iteration_graph.iteration_graph import IterationGraph
from . import ast
from ._exceptions import DiagonalAccessError, NoKernelFoundError
from ._to_iteration_graphs import to_iteration_graphs


def best_algorithm(
    assignment: ast.Assignment, formats: dict[str, Format]
) -> Result[IterationGraph, DiagonalAccessError | NoKernelFoundError]:
    try:
        match next(to_iteration_graphs(assignment, formats), None):
            case None:
                return Failure(NoKernelFoundError())
            case graph:
                return Success(graph)
    except DiagonalAccessError as e:
        return Failure(e)



__all__ = ["generate_module_tensora"]

from returns.result import Failure, Result, Success

from ..desugar import (
    DiagonalAccessError,
    NoKernelFoundError,
    best_algorithm,
    desugar_assignment,
    index_dimensions,
    to_identifiable,
)
from ..ir import peephole
from ..ir.ast import Module
from ..iteration_graph import Definition, generate_ir
from ..kernel_type import KernelType
from ..problem import Problem


def generate_module_tensora(
    problem: Problem, kernel_types: list[KernelType]
) -> Result[Module, DiagonalAccessError | NoKernelFoundError]:
    formats = problem.formats

    desugar = desugar_assignment(problem.assignment)

    output_variable = to_identifiable(desugar.target, formats)

    definition = Definition(output_variable, formats, index_dimensions(desugar))

    match best_algorithm(desugar, formats):
        case Failure() as result:
            return result
        case Success(graph):
            pass
        case _:
            raise NotImplementedError()

    functions = [generate_ir(definition, graph, kernel_type) for kernel_type in kernel_types]
    module = Module(functions)

    return Success(peephole(module))



__all__ = [
    "Expression",
    "Literal",
    "Integer",
    "Float",
    "Tensor",
    "Add",
    "Multiply",
    "Contract",
    "Assignment",
]

from dataclasses import dataclass


class Expression:
    __slots__ = ()


class Literal(Expression):
    __slots__ = ()


@dataclass(frozen=True, slots=True)
class Integer(Literal):
    value: int


@dataclass(frozen=True, slots=True)
class Float(Literal):
    value: float


@dataclass(frozen=True, slots=True)
class Tensor(Expression):
    id: int
    name: str
    indexes: tuple[str, ...]


@dataclass(frozen=True, slots=True)
class Add(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class Multiply(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True, slots=True)
class Contract(Expression):
    index: str
    expression: Expression


@dataclass(frozen=True, slots=True)
class Assignment:
    target: Tensor
    expression: Expression



__all__ = ["index_dimensions"]

from functools import singledispatch

from ..iteration_graph import TensorDimension
from .ast import Add, Assignment, Contract, Expression, Float, Integer, Multiply, Tensor


@singledispatch
def index_dimensions_expression(self: Expression) -> dict[str, TensorDimension]:
    raise NotImplementedError(f"index_dimensions not implemented for {type(self)}: {self}")


@index_dimensions_expression.register(Integer)
@index_dimensions_expression.register(Float)
def index_dimensions_nothing(self: Integer | Float) -> dict[str, TensorDimension]:
    return {}


@index_dimensions_expression.register(Tensor)
def index_dimensions_tensor(self: Tensor) -> dict[str, TensorDimension]:
    indexes = {}
    for i, index_i in enumerate(self.indexes):
        if index_i not in indexes:
            indexes[index_i] = TensorDimension(self.name, i)
    return indexes


@index_dimensions_expression.register(Add)
@index_dimensions_expression.register(Multiply)
def index_dimensions_add(self: Add | Multiply) -> dict[str, TensorDimension]:
    left_dimensions = index_dimensions_expression(self.left)
    right_dimensions = index_dimensions_expression(self.right)

    indexes = left_dimensions.copy()
    for index_i, dimension in right_dimensions.items():
        if index_i not in indexes:
            indexes[index_i] = dimension
    return indexes


@index_dimensions_expression.register(Contract)
def index_dimensions_contract(self: Contract) -> dict[str, TensorDimension]:
    return index_dimensions_expression(self.expression)


def index_dimensions(self: Assignment) -> dict[str, TensorDimension]:
    """Find a tensor name and dimension for each index in the assignment.

    The only way a kernel can know the size of an index is to get it from one
    of the tensors with a dimension indexed by that index. For each index, there
    will usually be multiple tensors whose dimension is indexed by that index,
    but they should all have the same size. This function returns the first one
    it finds for each index.
    """
    target_dimensions = index_dimensions_expression(self.target)
    right_dimensions = index_dimensions_expression(self.expression)

    indexes = target_dimensions.copy()
    for index_i, dimension in right_dimensions.items():
        if index_i not in indexes:
            indexes[index_i] = dimension
    return indexes



__all__ = ["AppendOutput"]

from dataclasses import dataclass

from ...format import Mode
from ...ir import SourceBuilder, types
from ...ir.ast import ArrayAllocate, Expression, IntegerLiteral, Multiply, Variable
from ...kernel_type import KernelType
from .._names import (
    crd_capacity_name,
    crd_name,
    dimension_name,
    layer_pointer,
    pos_capacity_name,
    pos_name,
    previous_layer_pointer,
    vals_capacity_name,
    vals_name,
)
from ..identifiable_expression import TensorLayer
from ..identifiable_expression import ast as ie_ast
from ._base import Output
from ._bucket import BucketOutput

default_array_size = Multiply(IntegerLiteral(1024), IntegerLiteral(1024))


@dataclass(frozen=True, slots=True)
class AppendOutput(Output):
    output: ie_ast.Tensor
    next_layer: int

    def vals_pointer(self) -> Expression:
        return previous_layer_pointer(self.output.id, self.output.order)

    def write_declarations(self, kernel_type: KernelType):
        source = SourceBuilder("Output initialization")

        target_name = self.output.name
        output_tensor = Variable(target_name)

        all_dense = True
        for i, mode in enumerate(self.output.modes):
            if mode == Mode.dense:
                pass
            elif mode == Mode.compressed:
                if kernel_type.is_assemble():
                    # How pos is handled depends on what the previous modes were
                    if all_dense:
                        # If the previous dimensions were all dense, then the size of pos in this dimension is fixed
                        pos_size = Multiply.join(
                            [dimension_name(self.output.indexes[i_prev]) for i_prev in range(i)]
                        ).plus(1)
                    else:
                        # Otherwise, the value will change, so provide a good default
                        pos_size = default_array_size
                    pos_capacity = pos_capacity_name(target_name, i)
                    pos_array = pos_name(target_name, i)
                    source.append(pos_capacity.declare(types.integer).assign(pos_size))
                    source.append(pos_array.assign(ArrayAllocate(types.integer, pos_capacity)))
                    source.append(pos_array.idx(0).assign(0))

                    # crd is always the same
                    crd_capacity = crd_capacity_name(target_name, i)
                    source.append(crd_capacity.declare(types.integer).assign(default_array_size))
                    source.append(
                        crd_name(target_name, i).assign(ArrayAllocate(types.integer, crd_capacity))
                    )

                # This is the only thing that get written when doing compute
                source.append(layer_pointer(self.output.id, i).declare(types.integer).assign(0))

                all_dense = False
            else:
                raise NotImplementedError()

        if kernel_type.is_assemble():
            if all_dense:
                vals_size = Multiply.join(
                    [output_tensor.attr("dimensions").idx(i) for i in range(self.output.order)]
                )
            else:
                vals_size = default_array_size
            vals_capacity = vals_capacity_name(target_name)
            source.append(vals_capacity.declare(types.integer).assign(vals_size))
            source.append(vals_name(target_name).assign(ArrayAllocate(types.float, vals_capacity)))

        return source

    def write_assignment(self, right_hand_side: Expression, kernel_type: KernelType):
        source = SourceBuilder()

        if self.next_layer != self.output.order:
            raise RuntimeError()

        source.append(vals_name(self.output.name).idx(self.vals_pointer()).assign(right_hand_side))

        return source

    def write_cleanup(self, kernel_type: KernelType):
        source = SourceBuilder(f"Assembling output tensor {self.output.name}")

        if kernel_type.is_assemble():
            target_name = self.output.name
            output_tensor = Variable(target_name)

            for i, mode in enumerate(self.output.modes):
                if mode == Mode.dense:
                    pass
                elif mode == Mode.compressed:
                    source.append(
                        output_tensor.attr("indices")
                        .idx(i)
                        .idx(0)
                        .assign(pos_name(target_name, i))
                    )
                    source.append(
                        output_tensor.attr("indices")
                        .idx(i)
                        .idx(1)
                        .assign(crd_name(target_name, i))
                    )
                else:
                    raise NotImplementedError()
            source.append(output_tensor.attr("vals").assign(vals_name(target_name)))

        return source

    def next_output(
        self, iteration_output: TensorLayer | None, kernel_type: KernelType
    ) -> tuple[Output, SourceBuilder, SourceBuilder]:
        if iteration_output is not None and self.next_layer == iteration_output.layer:
            return AppendOutput(self.output, self.next_layer + 1), SourceBuilder(), SourceBuilder()
        else:
            # No layer or wrong layer was encountered
            dense_only_remaining = all(
                mode == Mode.dense for mode in self.output.modes[self.next_layer :]
            )
            if dense_only_remaining:
                next_output = BucketOutput(
                    self.output, list(range(self.next_layer, len(self.output.modes)))
                )
                dimension_names = [
                    dimension_name(index) for index in self.output.indexes[self.next_layer :]
                ]
                bucket = vals_name(self.output.name).plus(
                    previous_layer_pointer(self.output.id, self.next_layer).times(
                        Multiply.join(dimension_names)
                    )
                )
                return next_output, next_output.write_declarations(bucket), SourceBuilder()
            else:
                raise NotImplementedError(
                    "Encountered a sparse output layer preceded by a contraction layer or a later "
                    "output layer. This requires a hash table to store intermediate outputs, "
                    "which is not currently implemented."
                )



__all__ = [
    "Type",
    "Integer",
    "integer",
    "Float",
    "float",
    "Tensor",
    "tensor",
    "Mode",
    "mode",
    "Pointer",
    "Array",
    "FixedArray",
]

from dataclasses import dataclass


class Type:
    __slots__ = ()


@dataclass(frozen=True, slots=True)
class Integer(Type):
    pass


integer = Integer()


@dataclass(frozen=True, slots=True)
class Float(Type):
    pass


float = Float()


@dataclass(frozen=True, slots=True)
class Tensor(Type):
    pass


tensor = Tensor()


@dataclass(frozen=True, slots=True)
class Mode(Type):
    pass


mode = Mode()


@dataclass(frozen=True, slots=True)
class Pointer(Type):
    target: Type


@dataclass(frozen=True, slots=True)
class Array(Type):
    element: Type


@dataclass(frozen=True, slots=True)
class FixedArray(Type):
    element: Type
    n: int



__all__ = [
    "write_sparse_initialization",
    "write_crd_assembly",
    "write_pos_assembly",
    "write_pos_allocation",
]

from ..format import Mode
from ..ir import SourceBuilder, types
from ..ir.ast import ArrayReallocate, GreaterThanOrEqual, Max, Multiply, Variable
from ._names import dimension_name
from .identifiable_expression import TensorLayer


def write_sparse_initialization(leaf: TensorLayer) -> SourceBuilder:
    source = SourceBuilder()

    index_variable = leaf.layer_pointer()
    start_index = leaf.previous_layer_pointer()
    end_index = leaf.sparse_end_name()
    pos_array = leaf.pos_name()

    source.append(index_variable.declare(types.integer).assign(pos_array.idx(start_index)))
    source.append(end_index.declare(types.integer).assign(pos_array.idx(start_index.plus(1))))

    return source


def write_crd_assembly(output: TensorLayer) -> SourceBuilder:
    source = SourceBuilder("crd assembly")

    pointer = output.layer_pointer()
    capacity = output.crd_capacity_name()
    crd = output.crd_name()
    loop_variable = Variable(output.tensor.indexes[output.layer])

    with source.branch(GreaterThanOrEqual(pointer, capacity)):
        source.append(capacity.assign(capacity.times(2)))
        source.append(crd.assign(ArrayReallocate(crd, types.integer, capacity)))

    source.append(crd.idx(pointer).assign(loop_variable))

    return source


def write_pos_assembly(output: TensorLayer) -> SourceBuilder:
    source = SourceBuilder("pos assembly")

    pointer = output.layer_pointer()
    pos = output.pos_name()
    previous_pointer = output.previous_layer_pointer()

    source.append(pos.idx(previous_pointer.plus(1)).assign(pointer))

    return source


def write_pos_allocation(output: TensorLayer) -> SourceBuilder:
    dense_dimensions = []
    for i_layer in range(output.layer + 1, output.tensor.order):
        index_variable_i = output.tensor.indexes[i_layer]
        mode_i = output.tensor.modes[i_layer]
        if mode_i == Mode.compressed:
            break
        dense_dimensions.append(dimension_name(index_variable_i))

    layer_being_allocated = output.layer + len(dense_dimensions) + 1
    if layer_being_allocated == len(output.tensor.indexes):
        comment = "vals allocation"
        capacity = output.vals_capacity_name()
        array = output.vals_name()
        type = types.float
        bonus = 0
    else:
        comment = "pos allocation for next sparse layer"
        target_leaf = TensorLayer(output.tensor, output.layer + len(dense_dimensions) + 1)
        capacity = target_leaf.pos_capacity_name()
        array = target_leaf.pos_name()
        type = types.integer
        bonus = 1  # pos array is 1 longer

    source = SourceBuilder(comment)

    # TODO: The minimum capacity formulas do not seem consistent, double check them
    if len(dense_dimensions) == 0:
        # Peephole optimization cannot figure out that doubling is always bigger with no dense dimensions, so the
        # dropping of max() must be done manually.
        minimum_capacity = output.layer_pointer().plus(bonus)
        with source.branch(GreaterThanOrEqual(minimum_capacity, capacity)):
            source.append(capacity.assign(capacity.times(2)))
            source.append(array.assign(ArrayReallocate(array, type, capacity)))
    else:
        minimum_capacity = (
            output.layer_pointer().plus(1).times(Multiply.join(dense_dimensions)).plus(bonus)
        )

        with source.branch(GreaterThanOrEqual(minimum_capacity, capacity)):
            source.append(capacity.assign(Max(capacity.times(2), minimum_capacity)))
            source.append(array.assign(ArrayReallocate(array, type, capacity)))

    return source



__all__ = ["generate_c_code_taco", "TacoError"]

import re
from dataclasses import dataclass

from returns.result import Failure, Result, Success

from ..kernel_type import KernelType
from ..problem import Problem
from ._deparse_to_taco import deparse_to_taco


@dataclass(frozen=True, slots=True)
class TacoError(Exception):
    message: str

    def __str__(self) -> str:
        return self.message


def generate_c_code_taco(
    problem: Problem, kernel_types: list[KernelType]
) -> Result[str, TacoError]:
    from tensora_taco import taco_cli

    formats = problem.formats

    expression_string = deparse_to_taco(problem.assignment)
    format_string_arguments = []
    for name, format in formats.items():
        if format.order != 0:  # Taco does not like formats for scalars
            mode_string = "".join(mode.character for mode in format.modes)
            ordering_string = ",".join(map(str, format.ordering))
            format_string_arguments.append(f"-f={name}:{mode_string}:{ordering_string}")

    kernel_type_arguments = []
    for kernel_type in kernel_types:
        match kernel_type:
            case KernelType.evaluate:
                kernel_type_arguments.append("-print-evaluate")
            case KernelType.compute:
                kernel_type_arguments.append("-print-compute")
            case KernelType.assemble:
                kernel_type_arguments.append("-print-assembly")

    taco_arguments = [
        expression_string,
        "-print-nocolor",
        *kernel_type_arguments,
        *format_string_arguments,
    ]

    match taco_cli(taco_arguments):
        case Success(code):
            declaration_regex = r"^int (evaluate|compute|assemble)\([^(]\)"

            argument_string = ", ".join(
                [f"taco_tensora_t *{name}" for name in problem.formats.keys()]
            )

            reordered_code = re.sub(
                declaration_regex, rf"int \1({argument_string})", code, flags=re.MULTILINE
            )

            return Success(reordered_code)
        case Failure(message):
            return Failure(TacoError(message))



# This folder cannot be nested until the import needs of the base packages and
# the cffi extra are more differentiated. Right now, the cffi extra needs only
# the setuptools package and only in Python 3.12.



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from typer.testing import CliRunner

from tensora.cli import app

runner = CliRunner()


def test_help():
    result = runner.invoke(app, ["--help"])

    assert result.exit_code == 0
    assert "Usage:" in result.stdout
    assert "Options" in result.stdout


def test_cli():
    result = runner.invoke(app, ["y(i) = A(i,j) * x(j)", "-f", "A:ds"])

    assert result.exit_code == 0
    assert result.stdout.startswith("int32_t compute(taco_tensor_t* restrict y,")


def test_multiple_kernels():
    result = runner.invoke(
        app,
        ["y(i) = A(i,j) * x(j)", "-t", "compute", "-t", "evaluate", "-t", "assemble"],
    )

    assert result.exit_code == 0
    assert "compute" in result.stdout
    assert "evaluate" in result.stdout
    assert "assemble" in result.stdout


def test_llvm_language():
    result = runner.invoke(app, ["y(i) = A(i,j) * x(j)", "-l", "llvm"])

    assert result.exit_code == 0
    assert "getelementptr" in result.stdout


def test_write_to_file():
    # Use a temporary directory instead of a temporary file to avoid issue on Windows
    # where the same file cannot be opened by two processes at the same time.
    with TemporaryDirectory() as tmpdir:
        file = Path(tmpdir) / "output.c"
        result = runner.invoke(app, ["y(i) = A(i,j) * x(j)", "-o", str(file)])

        assert result.exit_code == 0
        assert result.stdout == ""
        assert file.read_text().startswith("int32_t compute(taco_tensor_t* restrict y,")


@pytest.mark.parametrize(
    "command",
    [
        ["a(i) = b(i) +"],
        ["y(i) = A(i,j) * x(j)", "-f=ds"],
        ["y(i) = A(i,j) * x(j)", "-f=A:d1s2"],
        ["y(i) = A(i,j) * x(j)", "-f=A:ds", "-f=A:dd"],
        ["y(i) = A(i,j) * x(j)", "-f=A:d"],
        ["y(i) = A(i,j) * x(j)", "-f=B:ds"],
        ["a(i) = A(i,i)"],
        ["A(i,j) = B(i,j) + C(j,i)", "-f=A:ds", "-f=B:ds", "-f=C:ds"],
    ],
)
def test_bad_input(command):
    result = runner.invoke(app, command, catch_exceptions=False)

    assert result.exit_code == 1



from __future__ import annotations

__all__ = ["Mode", "Format"]

from dataclasses import dataclass
from enum import Enum


class Mode(Enum):
    # Manually map these to the entries in .compile.taco_type_header.taco_mode_t
    dense = (0, "d")
    compressed = (1, "s")

    def __init__(self, c_int: int, character: "str"):
        self.c_int = c_int
        self.character = character

    @staticmethod
    def from_c_int(value: int) -> Mode:
        for member in Mode:
            if member.value[0] == value:
                return member
        raise ValueError(f"No member of Mode has the integer value {value}")

    def __repr__(self) -> str:
        return f"Mode.{self.name}"


@dataclass(frozen=True, slots=True)
class Format:
    modes: tuple[Mode, ...]
    ordering: tuple[int, ...]

    def __post_init__(self):
        from ._exceptions import InvalidModeOrderingError

        if set(self.ordering) != set(range(len(self.modes))):
            raise InvalidModeOrderingError(self.modes, self.ordering)

    @property
    def order(self):
        return len(self.modes)

    def deparse(self):
        if self.ordering == tuple(range(self.order)):
            return "".join(mode.character for mode in self.modes)
        else:
            return "".join(
                mode.character + str(ordering)
                for mode, ordering in zip(self.modes, self.ordering, strict=True)
            )



__all__ = ["hoist_declarations"]

from functools import singledispatch

from ..ir.ast import (
    Assignment,
    Block,
    Branch,
    Declaration,
    DeclarationAssignment,
    Expression,
    FunctionDefinition,
    Loop,
    Return,
    Statement,
)
from ..ir.types import Type


@singledispatch
def hoist_declarations_statement(self: Statement) -> dict[str, Type]:
    raise NotImplementedError(
        f"hoist_declarations_statement not implemented for {type(self).__name__}"
    )


@hoist_declarations_statement.register
def hoist_declarations_expression(self: Expression) -> dict[str, Type]:
    return {}


@hoist_declarations_statement.register
def hoist_declarations_declaration(self: Declaration) -> dict[str, Type]:
    return {self.name.name: self.type}


@hoist_declarations_statement.register
def hoist_declarations_assignment(self: Assignment) -> dict[str, Type]:
    return {}


@hoist_declarations_statement.register
def hoist_declarations_declaration_assignment(self: DeclarationAssignment) -> dict[str, Type]:
    return {self.target.name.name: self.target.type}


@hoist_declarations_statement.register
def hoist_declarations_block(self: Block) -> dict[str, Type]:
    result = {}
    for s in self.statements:
        result.update(hoist_declarations_statement(s))
    return result


@hoist_declarations_statement.register
def hoist_declarations_branch(self: Branch) -> dict[str, Type]:
    result = {}
    result.update(hoist_declarations_statement(self.if_true))
    result.update(hoist_declarations_statement(self.if_false))
    return result


@hoist_declarations_statement.register
def hoist_declarations_loop(self: Loop) -> dict[str, Type]:
    return hoist_declarations_statement(self.body)


@hoist_declarations_statement.register
def hoist_declarations_return(self: Return) -> dict[str, Type]:
    return {}


def hoist_declarations(fn: FunctionDefinition) -> dict[str, Type]:
    return hoist_declarations_statement(fn.body)



__all__ = [
    "Problem",
    "make_problem",
    "IncorrectDimensionsError",
    "UndefinedReferenceError",
    "UnusedFormatError",
]

from dataclasses import dataclass

from returns.result import Failure, Result, Success

from .expression.ast import Assignment
from .format import Format, Mode


@dataclass(frozen=True, slots=True)
class IncorrectDimensionsError(Exception):
    name: str
    actual: int
    expected: int
    assignment: Assignment

    def __str__(self):
        return (
            f"Expected each reference in an assignment to have a number of indexes matching the "
            f"order of the corresponding format, but variable {self.name} referenced in "
            f"{self.assignment} indexes has order {self.actual} while its format has order {self.expected}"
        )


@dataclass(frozen=True, slots=True)
class UndefinedReferenceError(Exception):
    name: str
    assignment: Assignment
    formats: list[str]

    def __str__(self):
        return (
            f"Excepted each reference in an assignment to have a corresponding format, "
            f"but variable {self.name} referenced in {self.assignment} was not found among the "
            f"given formats {self.formats}"
        )


@dataclass(frozen=True, slots=True)
class UnusedFormatError(Exception):
    name: str
    assignment: Assignment

    def __str__(self):
        return (
            f"Expected each format to be referenced in the assignment, "
            f"but format {self.name} was not referenced in {self.assignment}"
        )


@dataclass(frozen=True, slots=True)
class Problem:
    assignment: Assignment
    formats: dict[str, Format]

    def __post_init__(self):
        # This intentionally allows for names in formats that are not referenced in the assignment.
        # The CLI and porcelain API will not allow this, but this is just as valid as defining a
        # function with unused parameters.

        tensor_orders = self.assignment.variable_orders()
        for name, order in tensor_orders.items():
            if name not in self.formats:
                raise UndefinedReferenceError(name, self.assignment, list(self.formats.keys()))
            elif order != self.formats[name].order:
                raise IncorrectDimensionsError(
                    name, self.formats[name].order, order, self.assignment
                )

    def __eq__(self, other: object):
        if isinstance(other, Problem):
            # Problems are only equal if the formats orders are equal
            return self.assignment == other.assignment and tuple(self.formats.items()) == tuple(
                other.formats.items()
            )
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash((self.assignment, tuple(self.formats.items())))


def make_problem(
    assignment: Assignment, formats: dict[str, Format]
) -> Result[Problem, UnusedFormatError | UndefinedReferenceError | IncorrectDimensionsError]:
    """Create a Problem while filling in default formats.

    This does three things that the `Problem` constructor does not do:
    1. It reorders the formats to match the order the tensors appear in the assignment.
    2. It fills in any missing formats with all dense modes.
    3. It raises an exception if there are formats not referenced in the assignment.
    """

    tensor_orders = assignment.variable_orders()

    for name in formats.keys():
        if name not in tensor_orders:
            return Failure(UnusedFormatError(name, assignment))

    new_formats = {}
    for name, order in tensor_orders.items():
        if name not in formats:
            new_formats[name] = Format(tuple([Mode.dense] * order), tuple(range(order)))
        else:
            new_formats[name] = formats[name]

    try:
        problem = Problem(assignment, new_formats)
    except (UndefinedReferenceError, IncorrectDimensionsError) as error:
        return Failure(error)

    return Success(problem)



__all__ = ["exhaust_tensor"]

from functools import singledispatch

from .ast import Add, Expression, Integer, Literal, Multiply, Tensor


@singledispatch
def exhaust_tensor(self, reference: str) -> Expression:
    raise NotImplementedError(f"exhaust_tensor not implemented for {type(self)}: {self}")


@exhaust_tensor.register(Literal)
def exhaust_tensor_literal(self: Literal, reference: str):
    return self


@exhaust_tensor.register(Tensor)
def exhaust_tensor_tensor(self: Tensor, reference: str):
    if self.id == reference:
        return Integer(0)
    else:
        return self


@exhaust_tensor.register(Add)
def exhaust_tensor_add(self: Add, reference: str):
    left_exhausted = exhaust_tensor(self.left, reference)
    right_exhausted = exhaust_tensor(self.right, reference)
    if left_exhausted is self.left and right_exhausted is self.right:
        # Short circuit when there are no changes
        return self
    elif left_exhausted == Integer(0):
        # Covers the case where both are exhausted
        return right_exhausted
    elif right_exhausted == Integer(0):
        return left_exhausted
    else:
        return Add(left_exhausted, right_exhausted)


@exhaust_tensor.register(Multiply)
def exhaust_tensor_multiply(self: Multiply, reference: str):
    left_exhausted = exhaust_tensor(self.left, reference)
    right_exhausted = exhaust_tensor(self.right, reference)
    if left_exhausted is self.left and right_exhausted is self.right:
        # Short circuit when there are no changes
        return self
    elif left_exhausted == Integer(0) or right_exhausted == Integer(0):
        return Integer(0)
    else:
        return Multiply(left_exhausted, right_exhausted)



from tensora import Tensor, evaluate


def test_sum_non_adjacent():
    b = Tensor.from_lol([1, 2, 3])
    c = Tensor.from_lol([[1, 3, 5], [2, 4, 6]])
    d = Tensor.from_lol([7, 8, 9])
    actual = evaluate("a(i) = b(i) + c(j,i) + d(i)", "d", b=b, c=c, d=d)
    expected = Tensor.from_lol([11, 17, 23])
    assert actual == expected



import pytest
from returns.result import Failure

from tensora import Format, Mode
from tensora.format import InvalidModeOrderingError, parse_format, parse_named_format

format_strings = [
    ("", Format((), ())),
    ("d", Format((Mode.dense,), (0,))),
    ("s", Format((Mode.compressed,), (0,))),
    ("ds", Format((Mode.dense, Mode.compressed), (0, 1))),
    ("sd", Format((Mode.compressed, Mode.dense), (0, 1))),
    ("d1s0", Format((Mode.dense, Mode.compressed), (1, 0))),
    ("d1s0s2", Format((Mode.dense, Mode.compressed, Mode.compressed), (1, 0, 2))),
]


@pytest.mark.parametrize(("string", "format"), format_strings)
def test_parse_format(string, format):
    actual = parse_format(string).unwrap()
    assert actual == format


@pytest.mark.parametrize(("string", "format"), format_strings)
def test_deparse_format(string, format):
    actual = format.deparse()
    assert actual == string


@pytest.mark.parametrize("string", ["df", "1d0s", "d0s", "d0s1s1", "d1s2s3", "d3d1d2"])
def test_parse_bad_format(string):
    actual = parse_format(string)
    assert isinstance(actual, Failure)


def test_parse_named_format():
    actual = parse_named_format("A:d1s0s2").unwrap()
    assert actual == ("A", Format((Mode.dense, Mode.compressed, Mode.compressed), (1, 0, 2)))


def test_parse_bad_named_format():
    actual = parse_named_format("d1s0s2s3")
    assert isinstance(actual, Failure)


@pytest.mark.parametrize("string", ["A:d0s", "A:d3d1d2"])
def test_parse_bad_ordering_in_named_format(string):
    actual = parse_named_format(string)
    assert isinstance(actual, Failure)


def test_format_attributes():
    format = Format((Mode.dense, Mode.compressed), (1, 0))

    assert format.order == 2
    assert format.modes[0] == Mode.dense


def test_mode_dense_attributes():
    mode_dense = Mode.from_c_int(0)
    assert mode_dense.c_int == 0
    assert mode_dense.character == "d"


def test_mode_sparse_attributes():
    mode_dense = Mode.from_c_int(1)
    assert mode_dense.c_int == 1
    assert mode_dense.character == "s"


def test_mode_from_illegal_int():
    with pytest.raises(ValueError, match="No member of Mode"):
        Mode.from_c_int(3)


def test_differing_sizes():
    with pytest.raises(InvalidModeOrderingError):
        _ = Format((Mode.dense,), (0, 1))



import pytest

from tensora import Tensor


@pytest.fixture
def a_ds():
    return Tensor.from_aos(
        [[1, 0], [0, 1], [1, 2]], [2.0, -2.0, 4.0], dimensions=(2, 3), format="ds"
    )


@pytest.fixture
def b_ds():
    return Tensor.from_aos(
        [[1, 1], [1, 2], [0, 2]], [-3.0, 4.0, 3.5], dimensions=(2, 3), format="ds"
    )


@pytest.fixture
def c_ds_add():
    return Tensor.from_aos(
        [[1, 0], [0, 1], [1, 2], [1, 1], [0, 2]],
        [2.0, -2.0, 8.0, -3.0, 3.5],
        dimensions=(2, 3),
        format="ds",
    )


def test_add_ds_ds(a_ds, b_ds, c_ds_add):
    actual = a_ds + b_ds

    assert actual == c_ds_add



from __future__ import annotations

__all__ = ["StableSet", "StableFrozenSet"]

from typing import AbstractSet, Hashable, Iterator, MutableSet, TypeVar

Element = TypeVar("Element", bound=Hashable, covariant=True)


class StableSet(MutableSet[Element]):
    def __init__(self, *items: Element):
        # Rely on stable dictionary
        self._items = {item: None for item in items}

    def __len__(self) -> int:
        return len(self._items)

    def __contains__(self, x: Element) -> bool:
        return x in self._items

    def __iter__(self) -> Iterator[Element]:
        return iter(self._items)

    def __reversed__(self):
        return StableSet(*reversed(self._items))

    def __or__(self, other: StableSet[Element]) -> StableSet[Element]:
        return StableSet(*self._items, *other._items)

    def __eq__(self, other: StableSet[Element]):
        if isinstance(other, StableSet):
            return self._items == other._items
        else:
            return NotImplemented

    def __repr__(self) -> str:
        return f"StableSet({', '.join(repr(item) for item in self._items)})"

    def add(self, element: Element, /) -> None:
        self._items[element] = None

    def discard(self, element: Element, /) -> None:
        del self[element]


class StableFrozenSet(AbstractSet[Element]):
    def __init__(self, *items: Element):
        used = set()
        unique_items = []
        for item in items:
            if item not in used:
                unique_items.append(item)
                used.add(item)
        self._items = tuple(unique_items)
        self._set = frozenset(items)

    def __len__(self) -> int:
        return len(self._items)

    def __contains__(self, x: Element) -> bool:
        return x in self._set

    def __iter__(self) -> Iterator[Element]:
        return iter(self._items)

    def __reversed__(self):
        return StableFrozenSet(*reversed(self._items))

    def __or__(self, other: StableFrozenSet[Element]) -> StableFrozenSet[Element]:
        return StableFrozenSet(*self._items, *other._items)

    def __eq__(self, other):
        if isinstance(other, StableFrozenSet):
            return self._set == other._set
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self._set)

    def __repr__(self) -> str:
        return f"StableFrozenSet({', '.join(repr(item) for item in self._items)})"



from __future__ import annotations

__all__ = [
    "dimension_name",
    "pos_name",
    "crd_name",
    "vals_name",
    "pos_capacity_name",
    "crd_capacity_name",
    "vals_capacity_name",
    "layer_pointer",
    "previous_layer_pointer",
    "sparse_end_name",
    "value_from_crd",
]


from ..ir.ast import Expression, IntegerLiteral, Variable


def dimension_name(index_variable: str) -> Variable:
    return Variable(f"{index_variable}_dim")


def pos_name(tensor: str, layer: int) -> Variable:
    return Variable(f"{tensor}_{layer}_pos")


def crd_name(tensor: str, layer: int) -> Variable:
    return Variable(f"{tensor}_{layer}_crd")


def vals_name(tensor: str) -> Variable:
    return Variable(f"{tensor}_vals")


def pos_capacity_name(tensor: str, layer: int) -> Variable:
    return Variable(f"{tensor}_{layer}_pos_capacity")


def crd_capacity_name(tensor: str, layer: int) -> Variable:
    return Variable(f"{tensor}_{layer}_crd_capacity")


def vals_capacity_name(tensor: str) -> Variable:
    return Variable(f"{tensor}_vals_capacity")


def layer_pointer(reference: str, layer: int) -> Variable:
    return Variable(f"p_{reference}_{layer}")


def previous_layer_pointer(reference: str, layer: int) -> Expression:
    if layer == 0:
        return IntegerLiteral(0)
    else:
        return layer_pointer(reference, layer - 1)


def sparse_end_name(reference: str, layer: int) -> Variable:
    return Variable(f"p_{reference}_{layer}_end")


def layer_begin_name(reference: str, layer: int) -> Variable:
    return Variable(f"p_{reference}_{layer}_begin")


def value_from_crd(reference: str, layer: int) -> Variable:
    return Variable(f"i_{reference}_{layer}")



__all__ = ["ir_to_c_statement", "ir_to_c_function_definition", "ir_to_c"]

from functools import singledispatch

from ..ir.ast import (
    Add,
    And,
    ArrayAllocate,
    ArrayIndex,
    ArrayReallocate,
    Assignment,
    AttributeAccess,
    Block,
    BooleanLiteral,
    BooleanToInteger,
    Branch,
    Declaration,
    DeclarationAssignment,
    Equal,
    Expression,
    FloatLiteral,
    FunctionDefinition,
    GreaterThan,
    GreaterThanOrEqual,
    IntegerLiteral,
    LessThan,
    LessThanOrEqual,
    Loop,
    Max,
    Min,
    Module,
    Multiply,
    NotEqual,
    Or,
    Return,
    Statement,
    Subtract,
    Variable,
)
from ._type_to_c import type_to_c


def parens(code: Expression, wrap_me: type | tuple[type, ...]):
    string = ir_to_c_expression(code)
    if isinstance(code, wrap_me):
        string = f"({string})"
    return string


def indent_lines(lines: list[str]) -> list[str]:
    return ["  " + line for line in lines]


@singledispatch
def ir_to_c_expression(self: Expression) -> str:
    raise NotImplementedError(f"ir_to_c_expression not implemented for {type(self)}: {self}")


@ir_to_c_expression.register(Variable)
def ir_to_c_variable(self: Variable) -> str:
    return self.name


@ir_to_c_expression.register(AttributeAccess)
def ir_to_c_attribute_access(self: AttributeAccess) -> str:
    return f"{ir_to_c_expression(self.target)}->{self.attribute}"


@ir_to_c_expression.register(ArrayIndex)
def ir_to_c_array_index(self: ArrayIndex) -> str:
    return f"{ir_to_c_expression(self.target)}[{ir_to_c_expression(self.index)}]"


@ir_to_c_expression.register(IntegerLiteral)
def ir_to_c_integer_literal(self: IntegerLiteral) -> str:
    return str(self.value)


@ir_to_c_expression.register(FloatLiteral)
def ir_to_c_float_literal(self: FloatLiteral) -> str:
    return str(self.value)


@ir_to_c_expression.register(BooleanLiteral)
def ir_to_c_boolean_literal(self: BooleanLiteral) -> str:
    return str(int(self.value))


@ir_to_c_expression.register(Add)
def ir_to_c_add(self: Add) -> str:
    return f"{ir_to_c_expression(self.left)} + {ir_to_c_expression(self.right)}"


@ir_to_c_expression.register(Subtract)
def ir_to_c_subtract(self: Subtract) -> str:
    # Subtract does not have the associative property so it needs parentheses around the right operand if it has the
    # same precedence.
    return f"{ir_to_c_expression(self.left)} - {parens(self.right, (Add, Subtract))}"


@ir_to_c_expression.register(Multiply)
def ir_to_c_multiply(self: Multiply) -> str:
    return f"{parens(self.left, (Add, Subtract))} * {parens(self.right, (Add, Subtract))}"


@ir_to_c_expression.register(Equal)
def ir_to_c_equal(self: Equal) -> str:
    return f"{ir_to_c_expression(self.left)} == {ir_to_c_expression(self.right)}"


@ir_to_c_expression.register(NotEqual)
def ir_to_c_not_equal(self: NotEqual) -> str:
    return f"{ir_to_c_expression(self.left)} != {ir_to_c_expression(self.right)}"


@ir_to_c_expression.register(GreaterThan)
def ir_to_c_greater_than(self: GreaterThan) -> str:
    return f"{ir_to_c_expression(self.left)} > {ir_to_c_expression(self.right)}"


@ir_to_c_expression.register(LessThan)
def ir_to_c_less_than(self: LessThan) -> str:
    return f"{ir_to_c_expression(self.left)} < {ir_to_c_expression(self.right)}"


@ir_to_c_expression.register(GreaterThanOrEqual)
def ir_to_c_greater_than_or_equal(self: GreaterThanOrEqual) -> str:
    return f"{ir_to_c_expression(self.left)} >= {ir_to_c_expression(self.right)}"


@ir_to_c_expression.register(LessThanOrEqual)
def ir_to_c_less_than_or_equal(self: LessThanOrEqual) -> str:
    return f"{ir_to_c_expression(self.left)} <= {ir_to_c_expression(self.right)}"


@ir_to_c_expression.register(And)
def ir_to_c_and(self: And) -> str:
    return f"{parens(self.left, Or)} && {parens(self.right, Or)}"


@ir_to_c_expression.register(Or)
def ir_to_c_or(self: Or) -> str:
    return f"{ir_to_c_expression(self.left)} || {ir_to_c_expression(self.right)}"


@ir_to_c_expression.register(Max)
def ir_to_c_max(self: Max) -> str:
    return f"TACO_MAX({ir_to_c_expression(self.left)}, {ir_to_c_expression(self.right)})"


@ir_to_c_expression.register(Min)
def ir_to_c_min(self: Min) -> str:
    return f"TACO_MIN({ir_to_c_expression(self.left)}, {ir_to_c_expression(self.right)})"


@ir_to_c_expression.register(BooleanToInteger)
def ir_to_c_boolean_to_integer(self: BooleanToInteger) -> str:
    return f"(int32_t)({ir_to_c_expression(self.expression)})"


@ir_to_c_expression.register(ArrayAllocate)
def ir_to_c_array_allocate(self: ArrayAllocate) -> str:
    return f"malloc(sizeof({type_to_c(self.element_type)}) * {parens(self.n_elements, (Add, Subtract))})"


@ir_to_c_expression.register(ArrayReallocate)
def ir_to_c_array_reallocate(self: ArrayReallocate) -> str:
    old = ir_to_c_expression(self.old)
    return f"realloc({old}, sizeof({type_to_c(self.element_type)}) * {parens(self.n_elements, (Add, Subtract))})"


def ir_to_c_declaration(self: Declaration) -> str:
    return type_to_c(self.type, self.name.name)


@singledispatch
def ir_to_c_statement(self: Statement) -> list[str]:
    raise NotImplementedError(f"ir_to_c_statement not implemented for {type(self)}: {self}")


@ir_to_c_statement.register(Expression)
def convert_expression_to_statement(self: Expression) -> list[str]:
    # Every expression can also be a statement; convert it here
    return [ir_to_c_expression(self) + ";"]


@ir_to_c_statement.register(Declaration)
def convert_declaration_to_statement(self: Declaration) -> list[str]:
    return [ir_to_c_declaration(self) + ";"]


@ir_to_c_statement.register(Assignment)
def ir_to_c_assignment(self: Assignment) -> list[str]:
    target = ir_to_c_expression(self.target)
    if isinstance(self.value, Add) and self.value.left == self.target:
        if self.value.right == IntegerLiteral(1):
            return [f"{target}++;"]
        else:
            return [f"{target} += {ir_to_c_expression(self.value.right)};"]
    elif isinstance(self.value, Subtract) and self.value.left == self.target:
        if self.value.right == IntegerLiteral(1):
            return [f"{target}--;"]
        else:
            return [f"{target} -= {ir_to_c_expression(self.value.right)};"]
    elif isinstance(self.value, Multiply) and self.value.left == self.target:
        return [f"{target} *= {ir_to_c_expression(self.value.right)};"]
    else:
        return [f"{target} = {ir_to_c_expression(self.value)};"]


@ir_to_c_statement.register(DeclarationAssignment)
def ir_to_c_declaration_assignment(self: DeclarationAssignment) -> list[str]:
    return [f"{ir_to_c_declaration(self.target)} = {ir_to_c_expression(self.value)};"]


@ir_to_c_statement.register(Block)
def ir_to_c_block(self: Block) -> list[str]:
    lines = []
    need_separator = False

    if self.comment is not None:
        # Add comment if it there is one
        lines.append(f"// {self.comment}")

    for statement in self.statements:
        if isinstance(statement, Block):
            if len(lines) > 0:
                # Add blank line if there are preceding lines
                lines.append("")
            lines.extend(ir_to_c_block(statement))
            need_separator = True
        else:
            if need_separator:
                # A block immediately preceded this line, so add the separator
                lines.append("")
                need_separator = False
            lines.extend(ir_to_c_statement(statement))

    return lines


@ir_to_c_statement.register(Branch)
def ir_to_c_branch(self: Branch) -> list[str]:
    if_true_lines = ir_to_c_statement(self.if_true)
    if_false_lines = ir_to_c_statement(self.if_false)

    lines = []
    lines.append(f"if ({ir_to_c_expression(self.condition)}) {{")
    lines.extend(indent_lines(if_true_lines))

    if isinstance(self.if_false, Branch):
        # Special case if-else chain to be put at the same level
        lines.append(f"}} else {if_false_lines[0]}")
        lines.extend(if_false_lines[1:])
    elif self.if_false == Block([]):
        # Special case empty if_false block to emit no else branch
        lines.append("}")
    else:
        lines.append("} else {")
        lines.extend(indent_lines(if_false_lines))
        lines.append("}")

    return lines


@ir_to_c_statement.register(Loop)
def ir_to_c_loop(self: Loop) -> list[str]:
    return [
        f"while ({ir_to_c_expression(self.condition)}) {{",
        *indent_lines(ir_to_c_statement(self.body)),
        "}",
    ]


@ir_to_c_statement.register(Return)
def ir_to_c_return(self: Return) -> list[str]:
    return [f"return {ir_to_c_expression(self.value)};"]


def ir_to_c_function_definition(self: FunctionDefinition) -> str:
    return_type_string = type_to_c(self.return_type)
    name_string = ir_to_c_expression(self.name)
    parameters_string = ", ".join(map(ir_to_c_declaration, self.parameters))

    lines = [
        f"{return_type_string} {name_string}({parameters_string}) {{",
        *indent_lines(ir_to_c_statement(self.body)),
        "}",
    ]

    return "\n".join(lines)


def ir_to_c(self: Module) -> str:
    return "\n\n".join((ir_to_c_function_definition(function) for function in self.definitions))



from hypothesis import given

from tensora.compile import BroadcastTargetIndexError, TensorMethod
from tensora.desugar import DiagonalAccessError, NoKernelFoundError

from .strategies import problem_and_tensors


@given(problem_and_tensors())
def test_generate_cannot_crash(problem_inputs):
    problem, input_tensors = problem_inputs

    try:
        method = TensorMethod(problem)
    except (BroadcastTargetIndexError, DiagonalAccessError, NoKernelFoundError):
        return

    _ = method(**input_tensors)



from __future__ import annotations

__all__ = [
    "Expression",
    "Literal",
    "Integer",
    "Float",
    "Tensor",
    "Add",
    "Subtract",
    "Multiply",
    "Assignment",
]

from abc import abstractmethod
from dataclasses import dataclass


class Expression:
    __slots__ = ()

    @abstractmethod
    def variables(self) -> dict[str, list[Tensor]]:
        raise NotImplementedError()

    @abstractmethod
    def deparse(self) -> str:
        """Convert the assignment back into a string."""
        raise NotImplementedError()

    @abstractmethod
    def index_participants(self) -> dict[str, set[tuple[str, int]]]:
        """Map of index name to tensors and dimensions it exists in.

        Returns:
            A mapping where each key is the string name of an index and each value is a sets of
            pairs. In each pair, the first element is the name of a tensor and the second element
            is the dimension in which that index appears.
        """
        raise NotImplementedError()

    def __str__(self):
        return self.deparse()


class Literal(Expression):
    __slots__ = ()

    def variables(self) -> dict[str, list[Tensor]]:
        return {}

    def index_participants(self) -> dict[str, set[tuple[str, int]]]:
        return {}


@dataclass(frozen=True, slots=True)
class Integer(Literal):
    value: int

    def deparse(self):
        return str(self.value)


@dataclass(frozen=True, slots=True)
class Float(Literal):
    value: float

    def deparse(self):
        return str(self.value)


@dataclass(frozen=True, slots=True)
class Tensor(Expression):
    name: str
    indexes: tuple[str, ...]

    @property
    def order(self):
        return len(self.indexes)

    def variables(self) -> dict[str, list[Tensor]]:
        return {self.name: [self]}

    def deparse(self):
        return self.name + "(" + ",".join(self.indexes) + ")"

    def index_participants(self) -> dict[str, set[tuple[str, int]]]:
        participants = {}
        for i, index_name in enumerate(self.indexes):
            participants[index_name] = participants.get(index_name, set()) | {(self.name, i)}
        return participants


def merge_index_participants(left: Expression, right: Expression):
    left_indexes = left.index_participants()
    right_indexes = right.index_participants()
    return {
        index_name: left_indexes.get(index_name, set()) | right_indexes.get(index_name, set())
        for index_name in {*left_indexes.keys(), *right_indexes.keys()}
    }


@dataclass(frozen=True, slots=True)
class Add(Expression):
    left: Expression
    right: Expression

    def variables(self) -> dict[str, list[Tensor]]:
        variables_mapping = self.left.variables().copy()
        for name, variables in self.right.variables().items():
            if name in variables_mapping:
                variables_mapping[name] = [*variables_mapping[name], *variables]
            else:
                variables_mapping[name] = variables
        return variables_mapping

    def deparse(self):
        left_string = self.left.deparse()

        right_string = self.right.deparse()
        if isinstance(self.right, (Add, Subtract)):
            # Preserve AST even though addition is associative.
            right_string = f"({right_string})"

        return left_string + " + " + right_string

    def index_participants(self) -> dict[str, set[tuple[str, int]]]:
        return merge_index_participants(self.left, self.right)


@dataclass(frozen=True, slots=True)
class Subtract(Expression):
    left: Expression
    right: Expression

    def variables(self) -> dict[str, list[Tensor]]:
        variables_mapping = self.left.variables().copy()
        for name, variables in self.right.variables().items():
            if name in variables_mapping:
                variables_mapping[name] = [*variables_mapping[name], *variables]
            else:
                variables_mapping[name] = variables
        return variables_mapping

    def deparse(self):
        left_string = self.left.deparse()

        right_string = self.right.deparse()
        if isinstance(self.right, (Add, Subtract)):
            right_string = f"({right_string})"

        return left_string + " - " + right_string

    def index_participants(self) -> dict[str, set[tuple[str, int]]]:
        return merge_index_participants(self.left, self.right)


@dataclass(frozen=True, slots=True)
class Multiply(Expression):
    left: Expression
    right: Expression

    def variables(self) -> dict[str, list[Tensor]]:
        variables_mapping = self.left.variables().copy()
        for name, variables in self.right.variables().items():
            if name in variables_mapping:
                variables_mapping[name] = [*variables_mapping[name], *variables]
            else:
                variables_mapping[name] = variables
        return variables_mapping

    def deparse(self):
        left_string = self.left.deparse()
        if isinstance(self.left, (Add, Subtract)):
            left_string = f"({left_string})"

        right_string = self.right.deparse()
        # Preserve AST even though multiplication is associative.
        if isinstance(self.right, (Add, Subtract, Multiply)):
            right_string = f"({right_string})"

        return f"{left_string} * {right_string}"

    def index_participants(self) -> dict[str, set[tuple[str, int]]]:
        return merge_index_participants(self.left, self.right)


@dataclass(frozen=True)
class Assignment:
    target: Tensor
    expression: Expression

    def __post_init__(self):
        from ._exceptions import (
            InconsistentDimensionsError,
            MutatingAssignmentError,
            NameConflictError,
        )

        target_name = self.target.name

        # Mutable containers
        variable_orders: dict[str, int] = {target_name: self.target.order}
        index_names = set(self.target.indexes)

        # Validate expression
        variables_mapping = self.expression.variables()
        for name, variables in variables_mapping.items():
            if name == target_name:
                raise MutatingAssignmentError(self)

            for variable in variables:
                index_names.update(variable.indexes)

            (first, *rest) = variables
            for variable in rest:
                if first.order != variable.order:
                    raise InconsistentDimensionsError(self, first, variable)

            variable_orders[name] = first.order

        # Detect name conflicts
        conflicted_names = index_names.intersection(variable_orders.keys())
        if len(conflicted_names) > 0:
            raise NameConflictError(conflicted_names.pop(), self)

        self._variable_orders: dict[str, int]
        object.__setattr__(self, "_variable_orders", variable_orders)

    def deparse(self) -> str:
        return self.target.deparse() + " = " + self.expression.deparse()

    def index_participants(self) -> dict[str, set[tuple[str, int]]]:
        return merge_index_participants(self.target, self.expression)

    def variable_orders(self) -> dict[str, int]:
        """Number of dimensions of each variable.

        Returns:
            A mapping where each key is the string name of a variable and each value is the number
            of dimensions that variable has.
        """

        return self._variable_orders

    def __str__(self) -> str:
        return self.deparse()



__all__ = ["KernelType"]

from enum import Enum


class KernelType(str, Enum):
    # Python 3.10 does not support StrEnum, so do it manually
    assemble = "assemble"
    compute = "compute"
    evaluate = "evaluate"

    def is_assemble(self):
        return self == KernelType.assemble or self == KernelType.evaluate

    def is_compute(self):
        return self == KernelType.compute or self == KernelType.evaluate

    def __str__(self) -> str:
        return self.name



__all__ = ["app"]

from pathlib import Path
from typing import Annotated, Optional

import typer
from parsita import ParseError
from returns.result import Failure, Success

from .expression import parse_assignment
from .format import parse_named_format
from .generate import Language, TensorCompiler, generate_code
from .kernel_type import KernelType
from .problem import make_problem

app = typer.Typer()


@app.command()
def tensora(
    assignment: Annotated[
        str,
        typer.Argument(
            show_default=False,
            help="The assignment for which to generate code, e.g. y(i) = A(i,j) * x(j).",
        ),
    ],
    target_format_strings: Annotated[
        list[str],
        typer.Option(
            "--format",
            "-f",
            help=(
                "A tensor and its format separated by a colon, e.g. A:d1s0 for CSC matrix. "
                "Unmentioned tensors are be assumed to be all dense."
            ),
        ),
    ] = [],  # noqa: B006; Typer does not support Sequence or tuple
    kernel_types: Annotated[
        list[KernelType],
        typer.Option(
            "--type",
            "-t",
            help="The type of kernel that will be generated. Can be mentioned multiple times.",
        ),
    ] = [KernelType.compute],  # noqa: B006; Typer does not support Sequence or tuple
    tensor_compiler: Annotated[
        TensorCompiler,
        typer.Option(
            "--compiler",
            "-c",
            help="The tensor algebra compiler to use to generate the kernel.",
        ),
    ] = TensorCompiler.tensora,
    language: Annotated[
        Language,
        typer.Option(
            "--language",
            "-l",
            help="The language in which to generate the kernel.",
        ),
    ] = Language.c,
    output_path: Annotated[
        Optional[Path],
        typer.Option(
            "--output",
            "-o",
            writable=True,
            help=(
                "The file to which the kernel will be written. If not specified, prints to "
                "standard out."
            ),
        ),
    ] = None,
):
    # Parse assignment
    match parse_assignment(assignment):
        case Failure(error):
            typer.echo(f"Failed to parse assignment:\n{error}", err=True)
            raise typer.Exit(1)
        case Success(parsed_assignment):
            pass
        case _:
            raise NotImplementedError()

    # Parse formats
    parsed_formats = {}
    for target_format_string in target_format_strings:
        match parse_named_format(target_format_string):
            case Failure(ParseError(_) as error):
                typer.echo(f"Failed to parse format:\n{error}", err=True)
                raise typer.Exit(1)
            case Failure(error):
                typer.echo(str(error), err=True)
                raise typer.Exit(1)
            case Success((target, format)):
                pass
            case _:
                raise NotImplementedError()

        if target in parsed_formats:
            typer.echo(f"Format for {target} was mentioned multiple times", err=True)
            raise typer.Exit(1)

        parsed_formats[target] = format

    # Validate and standardize assignment and formats
    match make_problem(parsed_assignment, parsed_formats):
        case Failure(error):
            typer.echo(str(error), err=True)
            raise typer.Exit(1)
        case Success(problem):
            pass
        case _:
            raise NotImplementedError()

    # Generate code
    match generate_code(problem, kernel_types, tensor_compiler, language):
        case Failure(error):
            typer.echo(str(error), err=True)
            raise typer.Exit(1)
        case Success(code):
            pass
        case _:
            raise NotImplementedError()

    if output_path is None:
        typer.echo(code)
    else:
        output_path.write_text(code)



import platform

from nox import Session, options, parametrize
from nox_uv import session

options.default_venv_backend = "uv"
options.sessions = ["test", "test_taco", "test_cffi", "test_numpy", "coverage", "lint"]


@session(python=["3.10", "3.11", "3.12", "3.13"], uv_groups=["test"])
def test(s: Session):
    coverage_file = f".coverage.{platform.machine()}.{platform.system()}.{s.python}"
    s.run("coverage", "run", "--data-file", coverage_file, "-m", "pytest", "tests")


@session(python=["3.10", "3.11", "3.12", "3.13"], uv_groups=["test"], uv_extras=["numpy"])
def test_numpy(s: Session):
    coverage_file = f".coverage.{platform.machine()}.{platform.system()}.{s.python}.numpy"
    s.run("coverage", "run", "--data-file", coverage_file, "-m", "pytest", "tests/test_numpy.py")


@session(python=["3.10", "3.11", "3.12", "3.13"], uv_groups=["test"], uv_extras=["cffi"])
def test_cffi(s: Session):
    coverage_file = f".coverage.{platform.machine()}.{platform.system()}.{s.python}.cffi"
    s.run("coverage", "run", "--data-file", coverage_file, "-m", "pytest", "tests_cffi")


@session(python=["3.10", "3.11", "3.12", "3.13"], uv_groups=["test"], uv_extras=["taco"])
def test_taco(s: Session):
    coverage_file = f".coverage.{platform.machine()}.{platform.system()}.{s.python}.taco"
    s.run("coverage", "run", "--data-file", coverage_file, "-m", "pytest", "tests/taco")


@session(venv_backend="none")
def coverage(s: Session):
    s.run("coverage", "combine")
    s.run("coverage", "html")
    s.run("coverage", "xml")


@session(venv_backend="none")
def fuzz(s: Session):
    s.run("hypothesis", "fuzz", "fuzz_tests")


@session(venv_backend="none")
@parametrize("command", [["ruff", "check", "."], ["ruff", "format", "--check", "."]])
def lint(s: Session, command: list[str]):
    s.run(*command)


@session(venv_backend="none")
def format(s: Session) -> None:
    s.run("ruff", "check", ".", "--select", "I", "--fix")
    s.run("ruff", "format", ".")



import pytest

from tensora import Tensor


@pytest.mark.parametrize(
    "array",
    [
        0.0,
        4.5,
        [],
        [[], []],
        [0, 0, 0],
        [[0, 1, 2], [0, 4, 5]],
        [[[0, 0, 3], [4, 5, 0]], [[0, 0, 0], [4, 5, 6]]],
    ],
)
@pytest.mark.parametrize("format", ["d", "s"])
def test_to_from_numpy(array, format):
    numpy = pytest.importorskip("numpy")

    expected = numpy.array(array)

    tensor = Tensor.from_numpy(expected, format=format * expected.ndim)
    actual = Tensor.to_numpy(tensor)

    assert numpy.array_equal(actual, expected)



__all__ = ["tensor_method", "evaluate", "evaluate_taco", "evaluate_cffi", "evaluate_tensora"]

from functools import lru_cache

from returns.functions import raise_exception

from ..expression import parse_assignment
from ..format import parse_format
from ..generate import TensorCompiler
from ..problem import Problem, make_problem
from ..tensor import Tensor
from ._tensor_method import BackendCompiler, TensorMethod


@lru_cache()
def cachable_tensor_method(
    problem: Problem, compiler: TensorCompiler, backend: BackendCompiler
) -> TensorMethod:
    return TensorMethod(problem, compiler, backend)


def tensor_method(
    assignment: str,
    formats: dict[str, str],
    compiler: TensorCompiler = TensorCompiler.tensora,
    backend: BackendCompiler | None = None,
) -> TensorMethod:
    parsed_assignment = parse_assignment(assignment).alt(raise_exception).unwrap()
    parsed_formats = {
        name: parse_format(format).alt(raise_exception).unwrap()
        for name, format in formats.items()
    }

    problem = make_problem(parsed_assignment, parsed_formats).alt(raise_exception).unwrap()

    if backend is None:
        if compiler == TensorCompiler.tensora:
            backend = BackendCompiler.llvm
        else:
            backend = BackendCompiler.cffi

    return cachable_tensor_method(problem, compiler, backend)


def evaluate_taco(assignment: str, output_format: str, **inputs: Tensor) -> Tensor:
    parsed_assignment = parse_assignment(assignment).alt(raise_exception).unwrap()
    input_formats = {name: tensor.format for name, tensor in inputs.items()}
    parsed_output_format = parse_format(output_format).alt(raise_exception).unwrap()

    formats = {parsed_assignment.target.name: parsed_output_format} | input_formats

    problem = make_problem(parsed_assignment, formats).alt(raise_exception).unwrap()

    function = cachable_tensor_method(problem, TensorCompiler.taco, BackendCompiler.cffi)

    return function(**inputs)


def evaluate_cffi(assignment: str, output_format: str, **inputs: Tensor) -> Tensor:
    parsed_assignment = parse_assignment(assignment).alt(raise_exception).unwrap()
    input_formats = {name: tensor.format for name, tensor in inputs.items()}
    parsed_output_format = parse_format(output_format).alt(raise_exception).unwrap()

    formats = {parsed_assignment.target.name: parsed_output_format} | input_formats

    problem = make_problem(parsed_assignment, formats).alt(raise_exception).unwrap()

    function = cachable_tensor_method(problem, TensorCompiler.tensora, BackendCompiler.cffi)

    return function(**inputs)


def evaluate_tensora(assignment: str, output_format: str, **inputs: Tensor) -> Tensor:
    parsed_assignment = parse_assignment(assignment).alt(raise_exception).unwrap()
    input_formats = {name: tensor.format for name, tensor in inputs.items()}
    parsed_output_format = parse_format(output_format).alt(raise_exception).unwrap()

    formats = {parsed_assignment.target.name: parsed_output_format} | input_formats

    problem = make_problem(parsed_assignment, formats).alt(raise_exception).unwrap()

    function = cachable_tensor_method(problem, TensorCompiler.tensora, BackendCompiler.llvm)

    return function(**inputs)


def evaluate(assignment: str, output_format: str, **inputs: Tensor) -> Tensor:
    return evaluate_tensora(assignment, output_format, **inputs)



from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest
from typer.testing import CliRunner

from tensora.cli import app

runner = CliRunner()


def test_multiple_kernels():
    result = runner.invoke(
        app,
        [
            "y(i) = A(i,j) * x(j)",
            "-t",
            "compute",
            "-t",
            "evaluate",
            "-t",
            "assemble",
            "-c",
            "taco",
        ],
    )

    assert result.exit_code == 0
    assert "compute" in result.stdout
    assert "evaluate" in result.stdout
    assert "assemble" in result.stdout


def test_write_to_file():
    with NamedTemporaryFile(suffix=".c") as f:
        result = runner.invoke(app, ["y(i) = A(i,j) * x(j)", "-o", f.name])

        assert result.exit_code == 0
        assert result.stdout == ""

        assert Path(f.name).read_text().startswith("int32_t compute(taco_tensor_t* restrict y,")


@pytest.mark.parametrize(
    "command",
    [
        ["a(i) = b(i) +"],
        ["y(i) = A(i,j) * x(j)", "-f=ds"],
        ["y(i) = A(i,j) * x(j)", "-f=A:d1s2"],
        ["y(i) = A(i,j) * x(j)", "-f=A:ds", "-f=A:dd"],
        ["y(i) = A(i,j) * x(j)", "-f=A:d"],
        ["y(i) = A(i,j) * x(j)", "-f=B:ds"],
        ["A(i,j) = B(i,j) + C(j,i)", "-f=A:ds", "-f=B:ds", "-f=C:ds"],
    ],
)
def test_bad_input(command):
    result = runner.invoke(app, [*command, "-c", "taco"], catch_exceptions=False)

    assert result.exit_code == 1
